#
# Copyright 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
#

import re
import urllib.parse

from .logger import log
from . import xmlutil


def sanitize_xml_for_test_define(xml):
    orig = xml
    xml = re.sub('arch=".*"', 'arch="i686"', xml)
    xml = re.sub('domain type=".*"', 'domain type="test"', xml)
    xml = re.sub('machine type=".*"', "", xml)
    xml = re.sub(">exe<", ">hvm<", xml)
    xml = re.sub(">linux<", ">xen<", xml)

    diff = xmlutil.diff(orig, xml)
    if diff:
        log.debug("virtinst test sanitizing diff\n:%s", diff)
    return xml


class URI:
    """
    Parse an arbitrary URI into its individual parts
    """

    def __init__(self, uri):
        self.uri = uri

        split_uri = self._split(uri)
        self.scheme = split_uri[0]
        (self.username, self.hostname, self.path, self.query, self.fragment) = map(
            urllib.parse.unquote, split_uri[1:]
        )

        self.transport = ""
        if "+" in self.scheme:
            self.scheme, self.transport = self.scheme.rsplit("+", 1)

        self.port = ""
        self.is_ipv6 = False
        if self.hostname.startswith("[") and "]" in self.hostname:
            if "]:" in self.hostname:
                self.hostname, self.port = self.hostname.rsplit(":", 1)
            self.hostname = "".join(self.hostname[1:].split("]", 1))
            self.is_ipv6 = True
        elif ":" in self.hostname:
            self.hostname, self.port = self.hostname.split(":", 1)

        self.host_is_ipv4_string = bool(re.match("^[0-9.]+$", self.hostname))

    ###################
    # Private helpers #
    ###################

    def _split(self, uri):
        def splitnetloc(url, start=0):
            for c in "/?#":  # the order is important!
                delim = url.find(c, start)
                if delim >= 0:
                    break
            else:
                delim = len(url)
            return url[start:delim], url[delim:]

        scheme = username = netloc = query = fragment = ""
        i = uri.find(":")
        if i > 0:
            scheme, uri = uri[:i].lower(), uri[i + 1 :]
            if uri[:2] == "//":
                netloc, uri = splitnetloc(uri, 2)
                offset = netloc.find("@")
                if offset > 0:
                    username = netloc[0:offset]
                    netloc = netloc[offset + 1 :]
            if "#" in uri:
                uri, fragment = uri.split("#", 1)
            if "?" in uri:
                uri, query = uri.split("?", 1)
        return scheme, username, netloc, uri, query, fragment


class MagicURI:
    """
    Handle magic virtinst URIs we use for the test suite and UI testing.
    This allows a special URI to override various features like capabilities
    XML, reported connection and libvirt versions, that enable testing
    different code paths.

    A magic URI has 3 parts:

        1) Magic prefix __virtinst_test__
        2) Actual openable URI, usually a test:/// URI
        3) Comma separated options

    The available options are:

        * 'predictable': Generate predictable UUIDs, MAC addresses, and
            temporary file names.
        * 'fakeuri': The URI to advertise as the actual connection URI
        * 'connver=%d': Override the connection (hv) version
        * 'libver=%d': Override the libvirt version
        * 'caps=%s': Points to a file with capabilities XML, that will
                     be returned in conn.getCapabilities. Ex.
                     files in test/capabilities-xml/
        * 'domcaps=%s': Points to a file with domain capabilities XML, that
                        will be returned in conn.getDomainCapabilities

    See tests/utils.py for example URLs
    """

    VIRTINST_URI_MAGIC_PREFIX = "__virtinst_test__"

    @staticmethod
    def uri_is_magic(uri):
        return uri.startswith(MagicURI.VIRTINST_URI_MAGIC_PREFIX)

    def __init__(self, uri):
        assert self.uri_is_magic(uri)

        from .cli import parse_optstr_tuples

        uri = uri.replace(self.VIRTINST_URI_MAGIC_PREFIX, "")
        ret = uri.split(",", 1)
        self.open_uri = ret[0]
        opts = dict(parse_optstr_tuples(len(ret) > 1 and ret[1] or ""))

        def pop_bool(field):
            ret = field in opts
            opts.pop(field, None)
            return ret

        self.predictable = pop_bool("predictable")
        self.fakeuri = opts.pop("fakeuri", None)
        self.capsfile = opts.pop("caps", None)
        self.domcapsfile = opts.pop("domcaps", None)

        self.conn_version = opts.pop("connver", None)
        if self.conn_version:
            self.conn_version = int(self.conn_version)
        elif self.fakeuri:
            self.conn_version = 10000000000

        self.libvirt_version = opts.pop("libver", None)
        if self.libvirt_version:
            self.libvirt_version = int(self.libvirt_version)

        self._err = None
        if opts:
            self._err = "MagicURI has unhandled opts=%s" % opts

    ##############
    # Public API #
    ##############

    def validate(self):
        if self._err:
            raise RuntimeError(self._err)

    def overwrite_conn_functions(self, conn):
        """
        After the connection is open, we need to stub out various functions
        depending on what magic bits the user specified in the URI
        """
        # Fake capabilities
        if self.capsfile:
            capsxml = open(self.capsfile).read()
            conn.getCapabilities = lambda: capsxml

        def _raise_nosupport_error(msg):
            import libvirt

            err = [libvirt.VIR_ERR_NO_SUPPORT, None, msg, None, None, None]
            exc = libvirt.libvirtError(msg)
            exc.err = err
            raise exc

        # Fake domcapabilities. This is insufficient since output should
        # vary per type/arch/emulator combo, but it can be expanded later
        # if needed
        domcapsxml = None
        if self.domcapsfile:
            domcapsxml = open(self.domcapsfile).read()

        def fake_domcaps(emulator, arch, machine, virttype, flags=0):
            ignore = emulator
            ignore = flags
            ignore = machine
            ignore = virttype

            if domcapsxml:
                ret = domcapsxml
                if arch:
                    ret = re.sub("arch>.+</arch", "arch>%s</arch" % arch, ret)
                return ret

            # In libvirt 9.8.0 the test suite added a stub domcaps
            # impl. Fall back to raising NO_SUPPORT for our magic URIs, so
            # we can keep getting code coverage of the old code paths
            _raise_nosupport_error("virtinst test driver fake domcaps nosupport")

        conn.getDomainCapabilities = fake_domcaps

        if self.fakeuri:
            origcreate = conn.createXML
            origdefine = conn.defineXML

            def newcreate(xml, flags):
                xml = sanitize_xml_for_test_define(xml)
                return origcreate(xml, flags)

            def newdefine(xml):
                xml = sanitize_xml_for_test_define(xml)
                return origdefine(xml)

            conn.createXML = newcreate
            conn.defineXML = newdefine



# Copyright (C) 2009, 2012-2014 Red Hat, Inc.
# Copyright (C) 2009 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import stat
import traceback

from gi.repository import Gtk
from gi.repository import Pango

import virtinst
from virtinst import log
from virtinst import xmlutil

from .asyncjob import vmmAsyncJob
from .baseclass import vmmGObjectUI
from .lib import uiutil

STORAGE_ROW_CONFIRM = 0
STORAGE_ROW_CANT_DELETE = 1
STORAGE_ROW_PATH = 2
STORAGE_ROW_TARGET = 3
STORAGE_ROW_ICON_SHOW = 4
STORAGE_ROW_ICON = 5
STORAGE_ROW_ICON_SIZE = 6
STORAGE_ROW_TOOLTIP = 7


class _vmmDeleteBase(vmmGObjectUI):
    """
    Base class for both types of VM/device storage deleting wizards
    """

    def __init__(self):
        vmmGObjectUI.__init__(self, "delete.ui", "vmm-delete")
        self.vm = None

        self.builder.connect_signals(
            {
                "on_vmm_delete_delete_event": self.close,
                "on_delete_cancel_clicked": self.close,
                "on_delete_ok_clicked": self._finish_clicked_cb,
                "on_delete_remove_storage_toggled": self._toggle_remove_storage,
            }
        )
        self.bind_escape_key_close()
        self._cleanup_on_app_close()

        self.topwin.set_title(self._get_dialog_title())
        self._init_state()

    def _init_state(self):
        _prepare_storage_list(self.widget("delete-storage-list"))

    def show(self, parent, vm):
        log.debug("Showing delete wizard")
        self._set_vm(vm)
        self._reset_state()
        self.topwin.set_transient_for(parent)
        self.topwin.present()

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing delete wizard")
        self.topwin.hide()
        self._set_vm(None)
        return 1

    def _cleanup(self):
        pass

    ##########################
    # Initialization methods #
    ##########################

    def _set_vm(self, newvm):
        oldvm = self.vm
        if oldvm:
            oldvm.conn.disconnect_by_obj(self)
        if newvm:
            newvm.conn.connect("vm-removed", self._vm_removed_cb)
        self.vm = newvm

    def _reset_state(self):
        # Set VM name or disk.target in title'
        text = self._get_dialog_text()

        title_str = "<span size='large'>%s</span>" % xmlutil.xml_escape(text)
        self.widget("header-label").set_markup(title_str)

        self.topwin.resize(1, 1)
        self.widget("delete-cancel").grab_focus()

        # Show warning message if VM is running
        vm_active = self._vm_active_status()
        uiutil.set_grid_row_visible(self.widget("delete-warn-running-vm-box"), vm_active)

        # Enable storage removal by default
        remove_storage_default = self._get_remove_storage_default()
        self.widget("delete-remove-storage").set_active(remove_storage_default)
        self.widget("delete-remove-storage").toggled()
        diskdatas = self._get_disk_datas()
        _populate_storage_list(self.widget("delete-storage-list"), self.vm, self.vm.conn, diskdatas)

    ################
    # UI listeners #
    ################

    def _finish_clicked_cb(self, src):
        self._finish()

    def _vm_removed_cb(self, _conn, vm):
        if self.vm == vm:
            self.close()

    def _toggle_remove_storage(self, src):
        dodel = src.get_active()
        uiutil.set_grid_row_visible(self.widget("delete-storage-scroll"), dodel)

    #########################
    # finish/delete methods #
    #########################

    def _get_paths_to_delete(self):
        del_list = self.widget("delete-storage-list")
        model = del_list.get_model()

        paths = []
        if self.widget("delete-remove-storage").get_active():
            for row in model:
                if not row[STORAGE_ROW_CANT_DELETE] and row[STORAGE_ROW_CONFIRM]:
                    paths.append(row[STORAGE_ROW_PATH])
        return paths

    def _delete_finished_cb(self, error, details):
        self.reset_finish_cursor()

        if error is not None:
            self.err.show_err(error, details=details)

        self.close()

    def _finish(self):
        paths = self._get_paths_to_delete()

        if paths:
            title = _("Are you sure you want to delete the storage?")
            message = _("The following paths will be deleted:\n\n%s") % "\n".join(paths)
            ret = self.err.chkbox_helper(
                self.config.get_confirm_delstorage,
                self.config.set_confirm_delstorage,
                text1=title,
                text2=message,
            )
            if not ret:
                return

        self.set_finish_cursor()

        if not self._remove_device(paths):
            # Don't delete storage if device removal failed
            self._delete_finished_cb(None, None)
            return

        title, text = self._get_progress_text(paths)

        progWin = vmmAsyncJob(
            self._async_delete,
            [self.vm, paths],
            self._delete_finished_cb,
            [],
            title,
            text,
            self.topwin,
        )
        progWin.run()
        self._set_vm(None)

    def _async_delete(self, asyncjob, vm, paths):
        errdata = None
        storage_errors = []

        try:
            self._destroy_vm(vm)

            conn = vm.conn.get_backend()
            meter = asyncjob.get_meter()
            storage_errors = self._async_delete_paths(paths, conn, meter)

            self._delete_vm(vm)
            vm.conn.schedule_priority_tick(pollvm=True)
        except Exception as e:  # pragma: no cover
            errdata = (
                (
                    _("Error deleting virtual machine '%(vm)s': %(error)s")
                    % {"vm": vm.get_name(), "error": str(e)}
                ),
                "".join(traceback.format_exc()),
            )

        if not storage_errors and not errdata:
            return

        storage_errstr = ""
        for errinfo in storage_errors:
            storage_errstr += "%s\n%s\n" % (errinfo[0], errinfo[1])

        # We had extra storage errors. If there was another error message,
        # errors to it. Otherwise, build the main error around them.
        if errdata:  # pragma: no cover
            error, details = errdata
            details += "\n\n"
            details += _("Additionally, there were errors removing certain storage devices: \n")
            details += storage_errstr
        else:
            error = _("Errors encountered while removing certain storage devices.")
            details = storage_errstr

        asyncjob.set_error(error, details)

    def _async_delete_paths(self, paths, conn, meter):
        storage_errors = []
        for path in paths:
            try:
                log.debug("Deleting path: %s", path)
                meter.start(_("Deleting path '%s'") % path, None)
                self._async_delete_path(conn, path, meter)
            except Exception as e:
                storage_errors.append((str(e), "".join(traceback.format_exc())))
            meter.end()
        return storage_errors

    def _async_delete_path(self, conn, path, ignore):
        try:
            vol = conn.storageVolLookupByPath(path)
        except Exception:
            vol = None
            log.debug("Path '%s' is not managed. Deleting locally", path)

        if vol:
            vol.delete(0)
        else:
            os.unlink(path)

    ################
    # Subclass API #
    ################

    def _get_dialog_title(self):
        raise NotImplementedError

    def _get_dialog_text(self):
        raise NotImplementedError

    def _get_progress_text(self, paths):
        raise NotImplementedError

    def _get_disk_datas(self):
        raise NotImplementedError

    def _vm_active_status(self):
        raise NotImplementedError

    def _delete_vm(self, vm):
        raise NotImplementedError

    def _remove_device(self, paths):
        raise NotImplementedError

    def _destroy_vm(self, vm):
        raise NotImplementedError

    def _get_remove_storage_default(self):
        raise NotImplementedError


class vmmDeleteDialog(_vmmDeleteBase):
    """
    Dialog for deleting a VM and optionally its storage
    """

    @classmethod
    def show_instance(cls, parentobj, vm):
        try:
            if not cls._instance:
                cls._instance = vmmDeleteDialog()
            cls._instance.show(parentobj.topwin, vm)
        except Exception as e:  # pragma: no cover
            parentobj.err.show_err(_("Error launching delete dialog: %s") % str(e))

    def _get_dialog_title(self):
        return _("Delete Virtual Machine")

    def _get_dialog_text(self):
        return _("Delete '%(vmname)s'") % {"vmname": self.vm.get_name()}

    def _get_progress_text(self, paths):
        if paths:
            title = (
                _("Deleting virtual machine '%s' and selected storage (this may take a while)")
                % self.vm.get_name()
            )
            text = title
        else:
            title = _("Deleting virtual machine '%s'") % self.vm.get_name()
            text = title
        return [title, text]

    def _get_remove_storage_default(self):
        return True

    def _get_disk_datas(self):
        return _build_diskdata_for_vm(self.vm)

    def _vm_active_status(self):
        vm_active = self.vm.is_active()
        return vm_active

    def _remove_device(self, paths):
        dummy = paths
        return True

    def _delete_vm(self, vm):
        if vm.is_persistent():
            log.debug("Removing VM '%s'", vm.get_name())
            vm.delete()

    def _destroy_vm(self, vm):
        if vm.is_active():
            log.debug("Forcing VM '%s' power off.", vm.get_name())
            vm.destroy()


class vmmDeleteStorage(_vmmDeleteBase):
    """
    Dialog for removing a disk device from a VM and optionally deleting
    its storage
    """

    @staticmethod
    def remove_devobj_internal(vm, err, devobj, deleting_storage=False):
        log.debug("Removing device: %s", devobj)

        # Define the change
        try:
            vm.remove_device(devobj)
        except Exception as e:  # pragma: no cover
            err.show_err(_("Error Removing Device: %s") % str(e))
            return

        # Try to hot remove
        detach_err = ()
        try:
            vm.detach_device(devobj)
        except Exception as e:
            log.debug("Device could not be hotUNplugged: %s", str(e))
            detach_err = (str(e), "".join(traceback.format_exc()))

        if not detach_err:
            return True

        msg = _("This change will take effect after the next guest shutdown.")
        if deleting_storage:
            msg += " "
            msg += _("Storage will not be deleted.")

        err.show_err(
            _("Device could not be removed from the running machine"),
            details=(detach_err[0] + "\n\n" + detach_err[1]),
            text2=msg,
            buttons=Gtk.ButtonsType.OK,
            dialog_type=Gtk.MessageType.INFO,
        )

    def __init__(self, disk):
        _vmmDeleteBase.__init__(self)
        self.disk = disk

    def _get_dialog_title(self):
        return _("Remove Disk Device")

    def _get_dialog_text(self):
        return _("Remove disk device '%(target)s'") % {"target": self.disk.target}

    def _get_progress_text(self, paths):
        if paths:
            title = (
                _("Removing disk device '%s' and selected storage (this may take a while)")
                % self.disk.target
            )
        else:
            title = _("Removing disk device '%s'") % self.disk.target
        text = title
        return [title, text]

    def _get_remove_storage_default(self):
        return False

    def _get_disk_datas(self):
        return [_DiskData.from_disk(self.disk)]

    def _vm_active_status(self):
        return False

    def _remove_device(self, paths):
        deleting_storage = bool(paths)
        return vmmDeleteStorage.remove_devobj_internal(
            self.vm, self.err, self.disk, deleting_storage=deleting_storage
        )

    def _delete_vm(self, vm):
        pass

    def _destroy_vm(self, vm):
        pass


###################
# UI init helpers #
###################


class _DiskData:
    """
    Helper class to contain all info we need to decide whether we
    should default to deleting a path
    """

    @staticmethod
    def from_disk(disk):
        """
        Build _DiskData from a DeviceDisk object
        """
        return _DiskData(
            disk.target,
            disk.get_source_path(),
            disk.read_only,
            disk.shareable,
            disk.device in ["cdrom", "floppy"],
        )

    def __init__(self, label, path, ro, shared, is_media):
        self.label = label
        self.path = path
        self.ro = ro
        self.shared = shared
        self.is_media = is_media


def _build_diskdata_for_vm(vm):
    """
    Return a list of _DiskData for all VM resources the app attempts to delete
    """
    diskdatas = []
    for disk in vm.xmlobj.devices.disk:
        diskdatas.append(_DiskData.from_disk(disk))

    diskdatas.append(_DiskData("kernel", vm.get_xmlobj().os.kernel, True, False, True))
    diskdatas.append(_DiskData("initrd", vm.get_xmlobj().os.initrd, True, False, True))
    diskdatas.append(_DiskData("dtb", vm.get_xmlobj().os.dtb, True, False, True))

    return diskdatas


def _populate_storage_list(storage_list, vm, conn, diskdatas):
    """
    Fill in the storage list UI from the passed list of _DiskData
    """
    model = storage_list.get_model()
    model.clear()

    for diskdata in diskdatas:
        if not diskdata.path:
            continue

        # There are a few pieces here
        # 1) Can we even delete the storage? If not, make the checkbox
        #    inconsistent. self.can_delete decides this for us, and if
        #    we can't delete, gives us a nice message to show the user
        #    for that row.
        #
        # 2) If we can delete, do we want to delete this storage by
        #    default? Reasons not to, are if the storage is marked
        #    readonly or shareable, or is in use by another VM.

        default = False
        definfo = None
        vol = conn.get_vol_by_path(diskdata.path)
        can_del, delinfo = _can_delete(conn, vol, diskdata.path)

        if can_del:
            default, definfo = _do_we_default(conn, vm.get_name(), vol, diskdata)

        info = None
        if not can_del:
            info = delinfo
        elif not default:
            info = definfo

        icon = "dialog-warning"
        icon_size = Gtk.IconSize.LARGE_TOOLBAR

        row = [
            default,
            not can_del,
            diskdata.path,
            diskdata.label,
            bool(info),
            icon,
            icon_size,
            info,
        ]
        model.append(row)


def _prepare_storage_list(storage_list):
    # Checkbox, deletable?, storage path, target (hda), icon name,
    # icon size, tooltip
    model = Gtk.ListStore(bool, bool, str, str, bool, str, int, str)
    storage_list.set_model(model)
    storage_list.set_tooltip_column(STORAGE_ROW_TOOLTIP)

    confirmCol = Gtk.TreeViewColumn()
    targetCol = Gtk.TreeViewColumn(_("Target"))
    infoCol = Gtk.TreeViewColumn()
    pathCol = Gtk.TreeViewColumn(_("Storage Path"))
    pathCol.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
    pathCol.set_expand(True)

    storage_list.append_column(confirmCol)
    storage_list.append_column(pathCol)
    storage_list.append_column(targetCol)
    storage_list.append_column(infoCol)

    chkbox = Gtk.CellRendererToggle()
    chkbox.connect("toggled", _storage_item_toggled, storage_list)
    confirmCol.pack_start(chkbox, False)

    def sensitive_cb(column, cell, model, _iter, data):
        row = model[_iter]
        inconsistent = row[STORAGE_ROW_CANT_DELETE]
        sensitive = not inconsistent
        active = row[STORAGE_ROW_CONFIRM]
        chk = column.get_cells()[0]
        chk.set_property("inconsistent", inconsistent)
        chk.set_property("active", active)
        chk.set_property("sensitive", sensitive)

    confirmCol.set_cell_data_func(chkbox, sensitive_cb)
    confirmCol.set_sort_column_id(STORAGE_ROW_CANT_DELETE)

    path_txt = Gtk.CellRendererText()
    pathCol.pack_start(path_txt, True)
    pathCol.add_attribute(path_txt, "text", STORAGE_ROW_PATH)
    pathCol.set_sort_column_id(STORAGE_ROW_PATH)
    path_txt.set_property("width-chars", 50)
    path_txt.set_property("ellipsize", Pango.EllipsizeMode.MIDDLE)

    target_txt = Gtk.CellRendererText()
    targetCol.pack_start(target_txt, False)
    targetCol.add_attribute(target_txt, "text", STORAGE_ROW_TARGET)
    targetCol.set_sort_column_id(STORAGE_ROW_TARGET)

    info_img = Gtk.CellRendererPixbuf()
    infoCol.pack_start(info_img, False)
    infoCol.add_attribute(info_img, "visible", STORAGE_ROW_ICON_SHOW)
    infoCol.add_attribute(info_img, "icon-name", STORAGE_ROW_ICON)
    infoCol.add_attribute(info_img, "stock-size", STORAGE_ROW_ICON_SIZE)
    infoCol.set_sort_column_id(STORAGE_ROW_ICON)


def _storage_item_toggled(src, index, storage_list):
    active = src.get_active()

    model = storage_list.get_model()
    model[index][STORAGE_ROW_CONFIRM] = not active


def _can_delete(conn, vol, path):
    """Is the passed path even deletable"""
    msg = None

    if vol:
        # Managed storage
        pool_type = vol.get_parent_pool().get_type()
        if pool_type == virtinst.StoragePool.TYPE_ISCSI:
            msg = _("Cannot delete iSCSI share.")  # pragma: no cover
        elif pool_type == virtinst.StoragePool.TYPE_SCSI:
            msg = _("Cannot delete SCSI device.")  # pragma: no cover
    else:
        if conn.is_remote():
            msg = _("Cannot delete unmanaged remote storage.")
        elif not os.path.exists(path):
            msg = _("Path does not exist.")
        elif not os.access(os.path.dirname(path), os.W_OK):
            msg = _("No write access to parent directory.")
        elif stat.S_ISBLK(os.stat(path)[stat.ST_MODE]):  # pragma: no cover
            msg = _("Cannot delete unmanaged block device.")

    can_delete = bool(not msg)
    return (can_delete, msg)


def _do_we_default(conn, vm_name, vol, diskdata):
    """Returns (do we delete by default?, info string if not)"""
    info = []

    if diskdata.ro:
        info.append(_("Storage is read-only."))
    elif not vol and not os.access(diskdata.path, os.W_OK):
        info.append(_("No write access to path."))

    if diskdata.shared:
        info.append(_("Storage is marked as shareable."))

    if not info and diskdata.is_media:
        info.append(_("Storage is a media device."))

    try:
        names = virtinst.DeviceDisk.path_in_use_by(conn.get_backend(), diskdata.path)

        if len(names) > 1:
            names.remove(vm_name)
            namestr = "\n- ".join(names)
            msg = _("Storage is in use by the following virtual machines")
            msg += "\n- " + namestr
            info.append(msg)
    except Exception as e:  # pragma: no cover
        log.exception("Failed checking disk conflict: %s", str(e))
        info.append(_("Failed to check disk usage conflict."))

    infostr = "\n".join(info)
    do_default = bool(not infostr)
    return (do_default, infostr)



# Copyright (C) 2006-2008, 2015 Red Hat, Inc.
# Copyright (C) 2006 Daniel P. Berrange <berrange@redhat.com>
# Copyright (C) 2010 Marc-Andre Lureau <marcandre.lureau@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from gi.repository import Gtk
from gi.repository import Gdk

from virtinst import log

from .serialcon import vmmSerialConsole
from .sshtunnels import ConnectionInfo
from .viewers import SpiceViewer, VNCViewer, SPICE_GTK_IMPORT_ERROR
from ..baseclass import vmmGObject, vmmGObjectUI
from ..lib.keyring import vmmKeyring


# console-pages IDs
(
    _CONSOLE_PAGE_UNAVAILABLE,
    _CONSOLE_PAGE_SERIAL,
    _CONSOLE_PAGE_GRAPHICS,
    _CONSOLE_PAGE_CONNECT,
) = range(4)

# console-gfx-pages IDs
(_GFX_PAGE_VIEWER, _GFX_PAGE_AUTH, _GFX_PAGE_UNAVAILABLE) = range(3)


class _TimedRevealer(vmmGObject):
    """
    Revealer for the fullscreen toolbar, with a bit of extra logic to
    hide/show based on mouse over
    """

    def __init__(self, toolbar):
        vmmGObject.__init__(self)

        self._in_fullscreen = False
        self._timeout_id = None

        self._revealer = Gtk.Revealer()
        self._revealer.add(toolbar)

        # Adding the revealer to the eventbox seems to ensure the
        # eventbox always has 1 invisible pixel showing at the top of the
        # screen, which we can use to grab the pointer event to show
        # the hidden toolbar.

        self._ebox = Gtk.EventBox()
        self._ebox.add(self._revealer)
        self._ebox.set_halign(Gtk.Align.CENTER)
        self._ebox.set_valign(Gtk.Align.START)
        self._ebox.show_all()

        self._ebox.connect("enter-notify-event", self._enter_notify)
        self._ebox.connect("leave-notify-event", self._enter_notify)

    def _cleanup(self):
        self._ebox.destroy()
        self._ebox = None
        self._revealer.destroy()
        self._revealer = None
        self._timeout_id = None

    def _enter_notify(self, ignore1, ignore2):
        x, y = self._ebox.get_pointer()
        alloc = self._ebox.get_allocation()
        entered = bool(x >= 0 and y >= 0 and x < alloc.width and y < alloc.height)

        if not self._in_fullscreen:
            return

        # Pointer exited the toolbar, and toolbar is revealed. Schedule
        # a timeout to close it, if one isn't already scheduled
        if not entered and self._revealer.get_reveal_child():
            self._schedule_unreveal_timeout(1000)
            return

        self._unregister_timeout()
        if entered and not self._revealer.get_reveal_child():
            self._revealer.set_reveal_child(True)

    def _schedule_unreveal_timeout(self, timeout):
        if self._timeout_id:
            return  # pragma: no cover

        def cb():
            self._revealer.set_reveal_child(False)
            self._timeout_id = None

        self._timeout_id = self.timeout_add(timeout, cb)

    def _unregister_timeout(self):
        if self._timeout_id:  # pragma: no cover
            self.remove_gobject_timeout(self._timeout_id)
            self._timeout_id = None

    def force_reveal(self, val):
        self._unregister_timeout()
        self._in_fullscreen = val
        self._revealer.set_reveal_child(val)
        self._schedule_unreveal_timeout(2000)

    def get_overlay_widget(self):
        return self._ebox


def build_keycombo_menu(on_send_key_fn):
    menu = Gtk.Menu()

    def make_item(accel, combo):
        name = Gtk.accelerator_get_label(*Gtk.accelerator_parse(accel))
        item = Gtk.MenuItem(name)
        item.connect("activate", on_send_key_fn, combo)

        menu.add(item)

    make_item("<Control><Alt>BackSpace", ["Control_L", "Alt_L", "BackSpace"])
    make_item("<Control><Alt>Delete", ["Control_L", "Alt_L", "Delete"])
    make_item("<Control><Alt><Shift>Escape", ["Control_L", "Alt_L", "Shift_L", "Escape"])
    menu.add(Gtk.SeparatorMenuItem())

    for i in range(1, 13):
        make_item("<Control><Alt>F%d" % i, ["Control_L", "Alt_L", "F%d" % i])
    menu.add(Gtk.SeparatorMenuItem())

    make_item("Print", ["Print"])

    menu.show_all()
    return menu


class vmmOverlayToolbar:
    def __init__(self, on_leave_fn, on_send_key_fn):
        self._send_key_button = None
        self._keycombo_menu = None
        self._toolbar = None

        self.timed_revealer = None
        self._init_ui(on_leave_fn, on_send_key_fn)

    def _init_ui(self, on_leave_fn, on_send_key_fn):
        self._keycombo_menu = build_keycombo_menu(on_send_key_fn)

        self._toolbar = Gtk.Toolbar()
        self._toolbar.set_show_arrow(False)
        self._toolbar.set_style(Gtk.ToolbarStyle.BOTH_HORIZ)
        self._toolbar.get_accessible().set_name("Fullscreen Toolbar")

        # Exit button
        button = Gtk.ToolButton()
        button.set_label(_("Leave Fullscreen"))
        button.set_icon_name("view-restore")
        button.set_tooltip_text(_("Leave fullscreen"))
        button.show()
        button.get_accessible().set_name("Fullscreen Exit")
        self._toolbar.add(button)
        button.connect("clicked", on_leave_fn)

        self._send_key_button = Gtk.ToolButton()
        self._send_key_button.set_icon_name("preferences-desktop-keyboard-shortcuts")
        self._send_key_button.set_tooltip_text(_("Send key combination"))
        self._send_key_button.show_all()
        self._send_key_button.connect("clicked", self._on_send_key_button_clicked_cb)
        self._send_key_button.get_accessible().set_name("Fullscreen Send Key")
        self._toolbar.add(self._send_key_button)

        self.timed_revealer = _TimedRevealer(self._toolbar)

    def _on_send_key_button_clicked_cb(self, src):
        event = Gtk.get_current_event()
        win = self._toolbar.get_window()
        rect = Gdk.Rectangle()

        rect.y = win.get_height()
        self._keycombo_menu.popup_at_rect(
            win, rect, Gdk.Gravity.NORTH_WEST, Gdk.Gravity.NORTH_WEST, event
        )

    def cleanup(self):
        self._keycombo_menu.destroy()
        self._keycombo_menu = None
        self._toolbar.destroy()
        self._toolbar = None
        self.timed_revealer.cleanup()
        self.timed_revealer = None


def _cant_embed_graphics(ginfo):
    if ginfo.gtype in ["vnc", "spice"]:
        return

    msg = _("Cannot display graphical console type '%s'") % ginfo.gtype
    return msg


class _ConsoleMenu(vmmGObject):
    """
    Helper class for building the text/graphical console menu list
    """

    def __init__(self, show_cb, toggled_cb):
        vmmGObject.__init__(self)
        self._menu = Gtk.Menu()
        self._menu.connect("show", show_cb)
        self._toggled_cb = toggled_cb

    def _cleanup(self):
        self._menu.destroy()
        self._menu = None
        self._toggled_cb = None

    ################
    # Internal API #
    ################

    def _build_serial_menu_items(self, vm):
        ret = []
        for dev in vmmSerialConsole.get_serialcon_devices(vm):
            if dev.DEVICE_TYPE == "console":
                label = _("Text Console %d") % (dev.get_xml_idx() + 1)
            else:
                label = _("Serial %d") % (dev.get_xml_idx() + 1)

            tooltip = vmmSerialConsole.can_connect(vm, dev)
            ret.append([label, dev, tooltip])

        if not ret:
            ret = [[_("No text console available"), None, None]]
        return ret

    def _build_graphical_menu_items(self, vm):

        from ..device.gfxdetails import vmmGraphicsDetails

        ret = []
        found_default = False
        for gdev in vm.xmlobj.devices.graphics:
            idx = gdev.get_xml_idx()
            ginfo = ConnectionInfo(vm.conn, gdev)

            label = (
                _("Graphical Console")
                + " "
                + vmmGraphicsDetails.graphics_pretty_type_simple(gdev.type)
            )
            if idx > 0:
                label += " %s" % (idx + 1)

            tooltip = _cant_embed_graphics(ginfo)
            if not tooltip:
                if not found_default:
                    found_default = True
                else:
                    tooltip = _("virt-manager does not support more than one graphical console")

            ret.append([label, ginfo, tooltip])

        if not ret:
            ret = [[_("Graphical console not configured for guest"), None, None]]
        return ret

    def _get_selected_menu_item(self):
        for child in self._menu.get_children():
            if hasattr(child, "get_active") and child.get_active():
                return child

    ##############
    # Public API #
    ##############

    def rebuild_menu(self, vm):
        olditem = self._get_selected_menu_item()
        oldlabel = olditem and olditem.get_label() or None

        # Clear menu
        for child in self._menu.get_children():
            self._menu.remove(child)

        graphics = self._build_graphical_menu_items(vm)
        serials = self._build_serial_menu_items(vm)

        # Use label == None to tell the loop to add a separator
        items = graphics + [[None, None, None]] + serials

        last_item = None
        for label, dev, tooltip in items:
            if label is None:
                self._menu.add(Gtk.SeparatorMenuItem())
                continue

            sensitive = bool(dev and not tooltip)
            if not sensitive and not tooltip:
                tooltip = label

            active = False
            if oldlabel is None and sensitive:
                # Select the first selectable option
                oldlabel = label
            if label == oldlabel:
                active = True

            item = Gtk.RadioMenuItem()
            if last_item is None:
                last_item = item
            else:
                item.join_group(last_item)

            item.set_label(label)
            item.set_active(active and sensitive)
            item.set_sensitive(sensitive)
            item.set_tooltip_text(tooltip or None)
            item.vmm_data = dev
            if sensitive:
                item.connect("toggled", self._toggled_cb)
            self._menu.add(item)

        self._menu.show_all()

    def activate_default(self):
        for child in self._menu.get_children():
            if child.get_sensitive() and hasattr(child, "toggled"):
                child.toggled()
                return True
        return False

    def get_selected(self):
        row = self._get_selected_menu_item()
        if not row:
            row = self._menu.get_children()[0]
        return row.get_label(), row.vmm_data, row.get_tooltip_text()

    def get_menu(self):
        return self._menu


class vmmConsolePages(vmmGObjectUI):
    """
    Handles all the complex UI handling dictated by the spice/vnc widgets
    """

    __gsignals__ = {
        "page-changed": (vmmGObjectUI.RUN_FIRST, None, []),
        "leave-fullscreen": (vmmGObjectUI.RUN_FIRST, None, []),
        "change-title": (vmmGObjectUI.RUN_FIRST, None, []),
    }

    def __init__(self, vm, builder, topwin):
        vmmGObjectUI.__init__(self, "console.ui", None, builder=builder, topwin=topwin)

        self.vm = vm
        self.top_box = self.widget("console-pages")
        self._pointer_is_grabbed = False

        # State for disabling modifiers when keyboard is grabbed
        self._accel_groups = Gtk.accel_groups_from_object(self.topwin)
        self._gtk_settings_accel = None
        self._gtk_settings_mnemonic = None

        # Initialize display widget
        self._viewer = None
        self._viewer_connect_clicked = False
        self._in_fullscreen = False

        # Fullscreen toolbar
        self._keycombo_menu = build_keycombo_menu(self._do_send_key)

        self._overlay_toolbar_fullscreen = vmmOverlayToolbar(
            on_leave_fn=self._leave_fullscreen, on_send_key_fn=self._do_send_key
        )
        self.widget("console-overlay").add_overlay(
            self._overlay_toolbar_fullscreen.timed_revealer.get_overlay_widget()
        )

        # When the gtk-vnc and spice-gtk widgets are in non-scaling mode, we
        # make them fill the whole window, and they paint the non-VM areas of
        # the viewer black. But when scaling is enabled, the viewer widget is
        # constrained. This change makes sure the non-VM portions in that case
        # are also colored black, rather than the default theme window color.
        self.widget("console-gfx-viewport").modify_bg(Gtk.StateType.NORMAL, Gdk.Color(0, 0, 0))

        self.widget("console-pages").set_show_tabs(False)
        self.widget("serial-pages").set_show_tabs(False)
        self.widget("console-gfx-pages").set_show_tabs(False)

        self._consolemenu = _ConsoleMenu(
            self._on_console_menu_show_cb, self._on_console_menu_toggled_cb
        )
        self._serial_consoles = []

        # Signals are added by vmmVMWindow. Don't use connect_signals here
        # or it changes will be overwritten

        self.builder.connect_signals(
            {
                "on_console_pages_switch_page": self._page_changed_cb,
                "on_console_auth_password_activate": self._auth_login_cb,
                "on_console_auth_login_clicked": self._auth_login_cb,
                "on_console_connect_button_clicked": self._connect_button_clicked_cb,
            }
        )

        self.widget("console-gfx-pages").connect("switch-page", self._page_changed_cb)

    def _cleanup(self):
        self.vm = None

        if self._viewer:
            self._viewer.cleanup()  # pragma: no cover
        self._viewer = None

        self._overlay_toolbar_fullscreen.cleanup()

        for serial in self._serial_consoles:
            serial.cleanup()
        self._serial_consoles = []

        self._consolemenu.cleanup()
        self._consolemenu = None

    #################
    # Internal APIs #
    #################

    def _disable_modifiers(self):
        if self._gtk_settings_accel is not None:
            return  # pragma: no cover

        for g in self._accel_groups:
            self.topwin.remove_accel_group(g)

        settings = Gtk.Settings.get_default()
        self._gtk_settings_accel = settings.get_property("gtk-menu-bar-accel")
        settings.set_property("gtk-menu-bar-accel", None)

        self._gtk_settings_mnemonic = settings.get_property("gtk-enable-mnemonics")
        settings.set_property("gtk-enable-mnemonics", False)

    def _enable_modifiers(self):
        if self._gtk_settings_accel is None:
            return

        settings = Gtk.Settings.get_default()
        settings.set_property("gtk-menu-bar-accel", self._gtk_settings_accel)
        self._gtk_settings_accel = None

        if self._gtk_settings_mnemonic is not None:
            settings.set_property("gtk-enable-mnemonics", self._gtk_settings_mnemonic)

        for g in self._accel_groups:
            self.topwin.add_accel_group(g)

    def _do_send_key(self, src, keys):
        ignore = src

        if keys is not None:
            self._viewer.console_send_keys(keys)

    ###########################
    # Resize and scaling APIs #
    ###########################

    def _viewer_get_resizeguest_tooltip(self):
        tooltip = ""
        if self._viewer:
            tooltip = self._viewer.console_get_resizeguest_warning()
        return tooltip or ""

    def _sync_resizeguest_with_display(self):
        if not self._viewer:
            return

        val = bool(self.vm.get_console_resizeguest())
        self._viewer.console_set_resizeguest(val)

    def _set_size_to_vm(self):
        if not self._viewer_is_visible():
            return  # pragma: no cover

        w, h = self._viewer.console_get_preferred_size()
        if w <= 0 or h <= 0:  # pragma: no cover
            log.debug("_set_size_to_vm but no valid sizing found")
            return

        top_w, top_h = self.topwin.get_size()
        viewer_alloc = self.widget("console-gfx-scroll").get_allocation()

        valw = w + (top_w - viewer_alloc.width)
        valh = h + (top_h - viewer_alloc.height)

        log.debug("_set_size_to_vm vm=(%s, %s) window=(%s, %s)", w, h, valw, valh)
        self.topwin.unmaximize()
        self.topwin.resize(valw, valh)

    ################
    # Scaling APIs #
    ################

    def _sync_scaling_with_display(self):
        if not self._viewer:
            return

        scale_type = self.vm.get_console_scaling()

        if scale_type == self.config.CONSOLE_SCALE_NEVER:
            self._viewer.console_set_scaling(False)
        elif scale_type == self.config.CONSOLE_SCALE_ALWAYS:
            self._viewer.console_set_scaling(True)
        elif scale_type == self.config.CONSOLE_SCALE_FULLSCREEN:
            self._viewer.console_set_scaling(self._in_fullscreen)

    ###################
    # Fullscreen APIs #
    ###################

    def _leave_fullscreen(self, ignore=None):
        self.emit("leave-fullscreen")

    def _change_fullscreen(self, do_fullscreen):
        if do_fullscreen:
            self._in_fullscreen = True
            self.topwin.fullscreen()
            self._overlay_toolbar_fullscreen.timed_revealer.force_reveal(True)
        else:
            self._in_fullscreen = False
            self._overlay_toolbar_fullscreen.timed_revealer.force_reveal(False)
            self.topwin.unfullscreen()

        self._sync_scaling_with_display()

    ##########################
    # State tracking methods #
    ##########################

    def _show_vm_status_unavailable(self):
        if self.vm.is_crashed():  # pragma: no cover
            self._activate_vm_unavailable_page(_("Guest has crashed."))
        else:
            self._activate_vm_unavailable_page(_("Guest is not running."))

    def _close_viewer(self):
        self._leave_fullscreen()
        self._viewer_connect_clicked = False

        for serial in self._serial_consoles:
            serial.close()

        if self._viewer is None:
            return
        self._viewer.console_remove_display_from_widget(self.widget("console-gfx-viewport"))
        self._viewer.cleanup()
        self._viewer = None
        log.debug("Viewer object cleaned up")

    def _refresh_vm_state(self):
        self._activate_default_console_page()

    ###########################
    # console page navigation #
    ###########################

    def _activate_gfx_unavailable_page(self, msg):
        self._close_viewer()
        self.widget("console-gfx-pages").set_current_page(_GFX_PAGE_UNAVAILABLE)
        if msg:
            self.widget("console-gfx-unavailable").set_label("<b>" + msg + "</b>")

    def _activate_vm_unavailable_page(self, msg):
        """
        This is the top level error page. We should only set it for very
        specific error cases, because when it is set and the VM is running
        we take that to mean we should attempt to connect to the default
        console.
        """
        self._close_viewer()
        self.widget("console-pages").set_current_page(_CONSOLE_PAGE_UNAVAILABLE)
        if msg:
            self.widget("console-unavailable").set_label("<b>" + msg + "</b>")
        self._activate_gfx_unavailable_page(msg)

    def _activate_auth_page(self, withPassword, withUsername):
        (pw, username) = vmmKeyring.get_instance().get_console_password(self.vm)

        self.widget("console-auth-password").set_visible(withPassword)
        self.widget("label-auth-password").set_visible(withPassword)

        self.widget("console-auth-username").set_visible(withUsername)
        self.widget("label-auth-username").set_visible(withUsername)

        self.widget("console-auth-username").set_text(username)
        self.widget("console-auth-password").set_text(pw)

        has_keyring = vmmKeyring.get_instance().is_available()
        remember = bool(withPassword and pw) or (withUsername and username)
        remember = has_keyring and remember
        self.widget("console-auth-remember").set_sensitive(has_keyring)
        self.widget("console-auth-remember").set_active(remember)

        self.widget("console-gfx-pages").set_current_page(_GFX_PAGE_AUTH)

        if withUsername:
            self.widget("console-auth-username").grab_focus()
        else:
            self.widget("console-auth-password").grab_focus()

    def _activate_gfx_viewer_page(self):
        self.widget("console-pages").set_current_page(_CONSOLE_PAGE_GRAPHICS)
        self.widget("console-gfx-pages").set_current_page(_GFX_PAGE_VIEWER)
        if self._viewer:
            self._viewer.console_grab_focus()

    def _activate_console_connect_page(self):
        self.widget("console-pages").set_current_page(_CONSOLE_PAGE_CONNECT)

    def _viewer_is_visible(self):
        is_visible = self.widget("console-pages").is_visible()
        cpage = self.widget("console-pages").get_current_page()
        gpage = self.widget("console-gfx-pages").get_current_page()

        return bool(
            is_visible
            and cpage == _CONSOLE_PAGE_GRAPHICS
            and gpage == _GFX_PAGE_VIEWER
            and self._viewer
            and self._viewer.console_is_open()
        )

    def _viewer_can_usb_redirect(self):
        return self._viewer_is_visible() and self._viewer.console_has_usb_redirection()

    #########################
    # Viewer login attempts #
    #########################

    def _init_viewer(self, ginfo, errmsg):
        if self._viewer or not self.is_visible():
            return

        if errmsg:
            log.debug("No acceptable graphics to connect to")
            self._activate_gfx_unavailable_page(errmsg)
            return

        if not self.vm.get_console_autoconnect() and not self._viewer_connect_clicked:
            self._activate_console_connect_page()
            return

        self._activate_gfx_unavailable_page(_("Connecting to graphical console for guest"))

        log.debug("Starting connect process for %s", ginfo.logstring())
        try:
            if ginfo.gtype == "vnc":
                viewer_class = VNCViewer
            elif ginfo.gtype == "spice":
                # We do this here and not in the embed check, since user
                # is probably expecting their spice console to work, so we
                # should show an explicit failure
                if SPICE_GTK_IMPORT_ERROR:
                    raise RuntimeError("Error opening SPICE console: %s" % SPICE_GTK_IMPORT_ERROR)
                viewer_class = SpiceViewer

            self._viewer = viewer_class(self.vm, ginfo)
            self._connect_viewer_signals()

            self._viewer.console_open()
        except Exception as e:
            log.exception("Error connecting to graphical console")
            self._activate_gfx_unavailable_page(_("Error connecting to graphical console:\n%s") % e)

    def _set_credentials(self, src_ignore=None):
        passwd = self.widget("console-auth-password")
        username = self.widget("console-auth-username")

        if passwd.get_visible():
            self._viewer.console_set_password(passwd.get_text())
        if username.get_visible():
            self._viewer.console_set_username(username.get_text())

        if self.widget("console-auth-remember").get_active():
            vmmKeyring.get_instance().set_console_password(
                self.vm, passwd.get_text(), username.get_text()
            )
        else:
            vmmKeyring.get_instance().del_console_password(self.vm)

    ##########################
    # Viewer signal handling #
    ##########################

    def _viewer_add_display_cb(self, _src, display):
        self.widget("console-gfx-viewport").add(display)

        # Sync initial settings
        self._sync_scaling_with_display()
        self._sync_resizeguest_with_display()

    def _pointer_grabbed_cb(self, _src):
        self._pointer_is_grabbed = True
        self.emit("change-title")

    def _pointer_ungrabbed_cb(self, _src):
        self._pointer_is_grabbed = False
        self.emit("change-title")

    def _viewer_keyboard_grab_cb(self, src):
        self._viewer_sync_modifiers()

    def _serial_focus_changed_cb(self, src, event):
        self._viewer_sync_modifiers()

    def _viewer_sync_modifiers(self):
        serial_has_focus = any([s.has_focus() for s in self._serial_consoles])
        viewer_keyboard_grab = self._viewer and self._viewer.console_has_keyboard_grab()

        if serial_has_focus or viewer_keyboard_grab:
            self._disable_modifiers()
        else:
            self._enable_modifiers()

    def _viewer_auth_error_cb(self, _src, errmsg, viewer_will_disconnect):
        errmsg = _("Viewer authentication error: %s") % errmsg
        self.err.val_err(errmsg)

        if viewer_will_disconnect:
            # GtkVNC will disconnect after an auth error, so lets do it for
            # them and re-init the viewer (which will be triggered by
            # _refresh_vm_state if needed)
            self._activate_vm_unavailable_page(errmsg)

        self._refresh_vm_state()

    def _viewer_need_auth_cb(self, _src, withPassword, withUsername):
        self._activate_auth_page(withPassword, withUsername)

    def _viewer_agent_connected_cb(self, _src):
        # Tell the vmwindow to trigger a state refresh, since
        # resizeguest setting depends on the agent value
        if self.widget("console-pages").is_visible():  # pragma: no cover
            self.emit("page-changed")

    def _viewer_usb_redirect_error_cb(self, _src, errstr):
        self.err.show_err(
            _("USB redirection error"), text2=str(errstr), modal=True
        )  # pragma: no cover

    def _viewer_disconnected_set_page(self, errdetails, ssherr):
        if self.vm.is_runable():  # pragma: no cover
            # Exit was probably for legitimate reasons
            self._show_vm_status_unavailable()
            return

        msg = _("Viewer was disconnected.")
        errmsg = ""
        if errdetails:
            errmsg += "\n" + errdetails
        if ssherr:
            log.debug("SSH tunnel error output: %s", ssherr)
            errmsg += "\n\n"
            errmsg += _("SSH tunnel error output: %s") % ssherr

        if errmsg:
            self._activate_gfx_unavailable_page(msg + errmsg)
            return

        # If no error message was reported, this isn't a clear graphics
        # error that should block reconnecting. So use the top level
        # 'VM unavailable' page which makes it easier for the user to
        # reconnect.
        self._activate_vm_unavailable_page(msg)

    def _viewer_disconnected_cb(self, _src, errdetails, ssherr):
        self._activate_gfx_unavailable_page(_("Viewer is disconnecting."))
        log.debug("Viewer disconnected cb")

        # Make sure modifiers are set correctly
        self._viewer_sync_modifiers()

        self._viewer_disconnected_set_page(errdetails, ssherr)

    def _viewer_connected_cb(self, _src):
        log.debug("Viewer connected cb")
        self._activate_gfx_viewer_page()

        # Make sure modifiers are set correctly
        self._viewer_sync_modifiers()

    def _connect_viewer_signals(self):
        self._viewer.connect("add-display-widget", self._viewer_add_display_cb)
        self._viewer.connect("pointer-grab", self._pointer_grabbed_cb)
        self._viewer.connect("pointer-ungrab", self._pointer_ungrabbed_cb)
        self._viewer.connect("keyboard-grab", self._viewer_keyboard_grab_cb)
        self._viewer.connect("keyboard-ungrab", self._viewer_keyboard_grab_cb)
        self._viewer.connect("connected", self._viewer_connected_cb)
        self._viewer.connect("disconnected", self._viewer_disconnected_cb)
        self._viewer.connect("auth-error", self._viewer_auth_error_cb)
        self._viewer.connect("need-auth", self._viewer_need_auth_cb)
        self._viewer.connect("agent-connected", self._viewer_agent_connected_cb)
        self._viewer.connect("usb-redirect-error", self._viewer_usb_redirect_error_cb)

    ##############################
    # Console list menu handling #
    ##############################

    def _console_menu_view_selected(self):
        name, dev, errmsg = self._consolemenu.get_selected()
        is_graphics = hasattr(dev, "gtype")

        if self.vm.is_runable():
            self._show_vm_status_unavailable()
            return

        if errmsg or not dev or is_graphics:
            self.widget("console-pages").set_current_page(_CONSOLE_PAGE_GRAPHICS)
            self.idle_add(self._init_viewer, dev, errmsg)
            return

        target_port = dev.get_xml_idx()
        serial = None
        for s in self._serial_consoles:
            if s.name == name:
                serial = s
                break

        if not serial:
            serial = vmmSerialConsole(self.vm, target_port, name)
            serial.set_focus_callbacks(self._serial_focus_changed_cb, self._serial_focus_changed_cb)

            title = Gtk.Label(label=name)
            self.widget("serial-pages").append_page(serial.get_box(), title)
            self._serial_consoles.append(serial)

        if not self.vm.get_console_autoconnect() and not self._viewer_connect_clicked:
            self._activate_console_connect_page()
            return

        serial.open_console()
        page_idx = self._serial_consoles.index(serial)
        self.widget("console-pages").set_current_page(_CONSOLE_PAGE_SERIAL)
        self.widget("serial-pages").set_current_page(page_idx)

    def _populate_console_menu(self):
        self._consolemenu.rebuild_menu(self.vm)

    def _toggle_first_console_menu_item(self):
        # We iterate through the 'console' menu and activate the first
        # valid entry... hacky but it works
        self._populate_console_menu()
        found = self._consolemenu.activate_default()
        if not found:
            # Calling this with dev=None will trigger _init_viewer
            # which shows some meaningful errors
            self._console_menu_view_selected()

    def _activate_default_console_page(self):
        if self.vm.is_runable():
            self._show_vm_status_unavailable()
            return

        viewer_initialized = self._viewer and self._viewer.console_is_open()
        if viewer_initialized:
            return

        cpage = self.widget("console-pages").get_current_page()
        if cpage != _CONSOLE_PAGE_UNAVAILABLE:
            return

        # If we are in this condition it should mean the VM was
        # just started, so connect to the default page
        self._toggle_first_console_menu_item()

    def _on_console_menu_toggled_cb(self, src):
        self._console_menu_view_selected()

    def _on_console_menu_show_cb(self, src):
        self._populate_console_menu()

    ################
    # UI listeners #
    ################

    def _auth_login_cb(self, src):
        self._set_credentials()

    def _connect_button_clicked_cb(self, src):
        self._viewer_connect_clicked = True
        self._console_menu_view_selected()

    def _page_changed_cb(self, src, origpage, newpage):
        # Hide the contents of all other pages, so they don't screw
        # up window sizing
        for i in range(src.get_n_pages()):
            src.get_nth_page(i).set_visible(i == newpage)

        # Dispatch the next bit in idle_add, so the UI size can change
        self.idle_emit("page-changed")

    ###########################
    # API used by vmmVMWindow #
    ###########################

    def vmwindow_viewer_can_usb_redirect(self):
        return self._viewer_can_usb_redirect()

    def vmwindow_viewer_get_usb_widget(self):
        return self._viewer.console_get_usb_widget()

    def vmwindow_viewer_get_pixbuf(self):
        return self._viewer.console_get_pixbuf()

    def vmwindow_close(self):
        return self._activate_vm_unavailable_page(_("Viewer window closed."))

    def vmwindow_get_title_message(self):
        if self._pointer_is_grabbed and self._viewer:
            keystr = self._viewer.console_get_grab_keys()
            return _("Press %s to release pointer.") % keystr

    def vmwindow_activate_default_console_page(self):
        return self._activate_default_console_page()

    def vmwindow_refresh_vm_state(self):
        return self._refresh_vm_state()

    def vmwindow_set_size_to_vm(self):
        return self._set_size_to_vm()

    def vmwindow_set_fullscreen(self, do_fullscreen):
        self._change_fullscreen(do_fullscreen)

    def vmwindow_get_keycombo_menu(self):
        return self._keycombo_menu

    def vmwindow_get_console_list_menu(self):
        return self._consolemenu.get_menu()

    def vmwindow_get_viewer_is_visible(self):
        return self._viewer_is_visible()

    def vmwindow_get_resizeguest_tooltip(self):
        return self._viewer_get_resizeguest_tooltip()

    def vmwindow_sync_scaling_with_display(self):
        return self._sync_scaling_with_display()

    def vmwindow_sync_resizeguest_with_display(self):
        return self._sync_resizeguest_with_display()



#
# Copyright 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .device import Device
from .char import CharSource
from ..xmlbuilder import XMLChildProperty, XMLProperty


class DeviceRng(Device):
    XML_NAME = "rng"

    TYPE_RANDOM = "random"
    TYPE_EGD = "egd"
    TYPE_BUILTIN = "builtin"

    model = XMLProperty("./@model")

    backend_model = XMLProperty("./backend/@model")
    backend_type = XMLProperty("./backend/@type")

    source = XMLChildProperty(CharSource, is_single=True, relative_xpath="./backend")

    rate_bytes = XMLProperty("./rate/@bytes")
    rate_period = XMLProperty("./rate/@period")

    device = XMLProperty("./backend[@model='random']")

    ##################
    # Default config #
    ##################

    def set_defaults(self, guest):
        if not self.model:
            self.model = "virtio"



#
# Copyright 2021 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
#

import sys

from . import _progresspriv


class Meter:
    """
    Meter class that hides the internals of the backend implementation
    from virtinst and friends
    """

    # Used by virt-manager subclass
    format_number = _progresspriv.format_number
    format_time = _progresspriv.format_time

    def __init__(self, quiet=False):
        self._text = None
        self._size = None
        self._total_read = 0
        if quiet:
            self._meter = _progresspriv.BaseMeter()
        else:
            self._meter = _progresspriv.TextMeter(output=sys.stdout)

    def start(self, text, size):
        self._text = text
        self._size = size
        self._total_read = 0
        self._meter.start(text, size)

    def update(self, new_total):
        self._total_read = new_total
        self._meter.update(new_total)

    def end(self):
        self._meter.end()


def make_meter(quiet):
    return Meter(quiet=quiet)


def ensure_meter(meter):
    if meter:
        return meter
    return make_meter(quiet=True)



#
# Copyright 2006-2009, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

from .device import Device
from ..logger import log
from ..xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty


def _validate_port(name, val):
    if val is None:
        return val
    val = int(val)

    if val < 5900 and val != -1:
        raise ValueError(_("%s must be above 5900, or -1 for auto allocation") % name)
    return val


class _GraphicsListen(XMLBuilder):
    XML_NAME = "listen"

    type = XMLProperty("./@type")
    address = XMLProperty("./@address")
    network = XMLProperty("./@network")
    socket = XMLProperty("./@socket")


class DeviceGraphics(Device):
    XML_NAME = "graphics"

    TYPE_SDL = "sdl"
    TYPE_VNC = "vnc"
    TYPE_RDP = "rdp"
    TYPE_SPICE = "spice"

    _XML_PROP_ORDER = [
        "_type",
        "gl",
        "_port",
        "_tlsPort",
        "autoport",
        "websocket",
        "keymap",
        "_listen",
        "passwd",
        "display",
        "xauth",
    ]

    keymap = XMLProperty("./@keymap")

    def _set_type(self, val):
        self._type = val
        # connected is valid for Spice and VNC, but VNC accepts only 'keep' and
        # libvirt errors out if there is any other value
        if val == "vnc" and self.connected != "keep":
            self.connected = None

    def _get_type(self):
        return self._type

    _type = XMLProperty("./@type")
    type = property(_get_type, _set_type)

    def _set_port(self, val):
        val = _validate_port("Port", val)
        self.autoport = self._get_default_autoport()
        self._port = val

    def _get_port(self):
        return self._port

    _port = XMLProperty("./@port", is_int=True)
    port = property(_get_port, _set_port)

    def _set_tlsport(self, val):
        val = _validate_port("TLS Port", val)
        self.autoport = self._get_default_autoport()
        self._tlsPort = val

    def _get_tlsport(self):
        return self._tlsPort

    _tlsPort = XMLProperty("./@tlsPort", is_int=True)
    tlsPort = property(_get_tlsport, _set_tlsport)

    autoport = XMLProperty("./@autoport", is_yesno=True)
    websocket = XMLProperty("./@websocket", is_int=True)

    xauth = XMLProperty("./@xauth")
    display = XMLProperty("./@display")

    def _set_listen(self, val):
        if val == "none":
            self._set_listen_none()
        elif val == "socket":
            self._remove_all_listens()
            obj = self.listens.add_new()
            obj.type = "socket"
        else:
            self._remove_all_listens()
            self._listen = val

    def _get_listen(self):
        return self._listen

    _listen = XMLProperty("./@listen")
    listen = property(_get_listen, _set_listen)

    passwd = XMLProperty("./@passwd")
    passwdValidTo = XMLProperty("./@passwdValidTo")
    socket = XMLProperty("./@socket")
    connected = XMLProperty("./@connected")
    defaultMode = XMLProperty("./@defaultMode")

    listens = XMLChildProperty(_GraphicsListen)

    def _remove_all_listens(self):
        for listen in self.listens:
            self.remove_child(listen)

    def get_first_listen_type(self):
        if len(self.listens) > 0:
            return self.listens[0].type
        return None

    def _set_listen_none(self):
        self._remove_all_listens()
        self.listen = None
        self.port = None
        self.tlsPort = None
        self.autoport = None
        self.socket = None

        if self.conn.support.conn_graphics_listen_none():
            obj = self.listens.add_new()
            obj.type = "none"

    # Spice bits
    image_compression = XMLProperty("./image/@compression")
    streaming_mode = XMLProperty("./streaming/@mode")
    clipboard_copypaste = XMLProperty("./clipboard/@copypaste", is_yesno=True)
    mouse_mode = XMLProperty("./mouse/@mode")
    filetransfer_enable = XMLProperty("./filetransfer/@enable", is_yesno=True)
    gl = XMLProperty("./gl/@enable", is_yesno=True)
    rendernode = XMLProperty("./gl/@rendernode")
    zlib_compression = XMLProperty("./zlib/@compression")

    ##################
    # Default config #
    ##################

    def _listen_need_port(self):
        listen = self.get_first_listen_type()
        return not listen or listen in ["address", "network"]

    def _get_default_port(self):
        if self.type in ["vnc", "spice"] and self._listen_need_port():
            return -1
        return None

    def _get_default_tlsport(self):
        if self.type == "spice" and self._listen_need_port():
            return -1
        return None

    def _get_default_autoport(self):
        # By default, don't do this for VNC to maintain back compat with
        # old libvirt that didn't support 'autoport'
        if self.type != "spice":
            return None
        if self.port == -1 and self.tlsPort == -1:
            return True
        return None

    def _default_type(self, guest):
        gtype = guest.default_graphics_type
        log.debug("App is configured with default_graphics=%s", gtype)

        if self.conn.is_xen():
            # Xen domcaps can advertise spice support, but we have historically
            # not defaulted to it for xen, so force vnc.
            log.debug("Not defaulting to spice for xen driver. Using vnc.")
            gtype = "vnc"

        if gtype == "spice" and not guest.lookup_domcaps().supports_graphics_spice():
            log.debug("spice requested but HV doesn't support it. Using vnc.")
            gtype = "vnc"
        return gtype

    def _default_image_compression(self, _guest):
        if self.type != "spice":
            return None
        if not self.conn.is_remote():
            log.debug("Local connection, disabling spice image compression.")
            return "off"
        return None

    def _default_spice_gl(self, _guest):
        # If spice GL but rendernode wasn't specified, hardcode
        # the first one
        if not self.rendernode:
            for nodedev in self.conn.fetch_all_nodedevs():
                if not nodedev.is_drm_render():
                    continue
                self.rendernode = nodedev.get_devnode().path
                break

    def set_defaults(self, guest):
        if not self.type:
            self.type = self._default_type(guest)

        if self.type == "sdl":
            if not self.xauth:
                self.xauth = os.path.expanduser("~/.Xauthority")
            if not self.display:
                self.display = os.environ.get("DISPLAY")

        if self.port is None:
            self.port = self._get_default_port()
        if self.tlsPort is None:
            self.tlsPort = self._get_default_tlsport()
        if self.autoport is None:
            self.autoport = self._get_default_autoport()

        if not self.image_compression:
            self.image_compression = self._default_image_compression(guest)
        if self.type == "spice" and self.gl:
            self._default_spice_gl(guest)



#
# Copyright 2013-2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .devices.disk import _DiskSource
from .xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty


class _SnapshotDisk(XMLBuilder):
    XML_NAME = "disk"
    name = XMLProperty("./@name")
    snapshot = XMLProperty("./@snapshot")
    source = XMLChildProperty(_DiskSource, is_single=True)


class DomainSnapshot(XMLBuilder):
    XML_NAME = "domainsnapshot"
    _XML_PROP_ORDER = ["name", "description", "creationTime"]

    name = XMLProperty("./name")
    description = XMLProperty("./description")
    state = XMLProperty("./state")
    creationTime = XMLProperty("./creationTime", is_int=True)
    parent = XMLProperty("./parent/name")

    memory_type = XMLProperty("./memory/@snapshot")
    memory_file = XMLProperty("./memory/@file")

    disks = XMLChildProperty(_SnapshotDisk, relative_xpath="./disks")



#
# Copyright 2011, 2013 Red Hat, Inc.
# Copyright 2013 IBM Corporation
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .device import Device
from ..xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty


class _ActivePCRBanks(XMLBuilder):
    XML_NAME = "active_pcr_banks"

    sha1 = XMLProperty("./sha1", is_bool=True)
    sha256 = XMLProperty("./sha256", is_bool=True)
    sha384 = XMLProperty("./sha384", is_bool=True)
    sha512 = XMLProperty("./sha512", is_bool=True)


class DeviceTpm(Device):
    XML_NAME = "tpm"

    VERSION_1_2 = "1.2"
    VERSION_2_0 = "2.0"
    VERSIONS = [VERSION_1_2, VERSION_2_0]

    TYPE_PASSTHROUGH = "passthrough"
    TYPE_EMULATOR = "emulator"
    TYPES = [TYPE_PASSTHROUGH, TYPE_EMULATOR]

    MODEL_TIS = "tpm-tis"
    MODEL_CRB = "tpm-crb"
    MODEL_SPAPR = "tpm-spapr"
    MODELS = [MODEL_TIS, MODEL_CRB, MODEL_SPAPR]

    type = XMLProperty("./backend/@type")
    version = XMLProperty("./backend/@version")
    model = XMLProperty("./@model")
    device_path = XMLProperty("./backend/device/@path")
    encryption_secret = XMLProperty("./backend/encryption/@secret")
    persistent_state = XMLProperty("./backend/@persistent_state", is_yesno=True)
    debug = XMLProperty("./backend/@debug")
    source_type = XMLProperty("./backend/source/@type")
    source_path = XMLProperty("./backend/source/@path")
    profile_source = XMLProperty("./backend/profile/@source")
    profile_removeDisabled = XMLProperty("./backend/profile/@removeDisabled")

    active_pcr_banks = XMLChildProperty(_ActivePCRBanks, is_single=True, relative_xpath="./backend")

    ##################
    # Default config #
    ##################

    @staticmethod
    def default_model(guest):
        domcaps = guest.lookup_domcaps()

        if not domcaps.devices.tpm.present and not guest.os.is_pseries():
            # Preserve the old default when domcaps is old
            return DeviceTpm.MODEL_CRB
        if domcaps.devices.tpm.get_enum("model").has_value(DeviceTpm.MODEL_CRB):
            # CRB is the modern version, and it implies version 2.0
            return DeviceTpm.MODEL_CRB

        # Let libvirt decide so we don't need to duplicate its arch logic
        return None

    def set_defaults(self, guest):
        if self.device_path and not self.type:
            self.type = self.TYPE_PASSTHROUGH
        if not self.type:
            # Libvirt requires a backend type to be specified. 'emulator'
            # may not be available if swtpm is not installed, but trying to
            # fallback to 'passthrough' in that case isn't really workable.
            # Instead we specify it unconditionally and let libvirt error.
            self.type = self.TYPE_EMULATOR

        # passthrough and model and version are all interconnected, so
        # don't try to set a default model if other bits are set
        if self.type == self.TYPE_EMULATOR and not self.model and not self.version:
            self.model = self.default_model(guest)



#
# Copyright 2006-2009, 2013, 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

from . import urldetect
from . import urlfetcher
from .installerinject import perform_initrd_injections
from .. import progress
from ..devices import DeviceDisk
from ..logger import log
from ..osdict import OSDB


# Enum of the various install media types we can have
(MEDIA_DIR, MEDIA_ISO, MEDIA_URL, MEDIA_KERNEL) = range(1, 5)


def _is_url(url):
    return url.startswith("http://") or url.startswith("https://") or url.startswith("ftp://")


class _LocationData:
    def __init__(self, osinfo, kernel_pairs, os_media, os_tree):
        self.osinfo = osinfo
        self.kernel_pairs = kernel_pairs
        self.os_media = os_media
        self.os_tree = os_tree

        self.kernel_url_arg = None
        if self.osinfo:
            osobj = OSDB.lookup_os(self.osinfo)
            self.kernel_url_arg = osobj.get_kernel_url_arg()


class InstallerTreeMedia:
    """
    Class representing --location Tree media. Can be one of

      - A network URL: http://dl.fedoraproject.org/...
      - A local directory
      - A local .iso file, which will be accessed with isoinfo
    """

    @staticmethod
    def validate_path(conn, path):
        try:
            dev = DeviceDisk(conn)
            dev.device = dev.DEVICE_CDROM
            dev.set_source_path(path)
            dev.validate()
            return dev.get_source_path()
        except Exception as e:
            log.debug("Error validating install location", exc_info=True)
            if path.startswith("nfs:"):
                log.warning(
                    "NFS URL installs are no longer supported. "
                    "Access your install media over an alternate transport "
                    "like HTTP, or manually mount the NFS share and install "
                    "from the local directory mount point."
                )

            msg = _("Validating install media '%(media)s' failed: %(error)s") % {
                "media": str(path),
                "error": str(e),
            }
            raise ValueError(msg) from None

    @staticmethod
    def get_system_scratchdir(guest):
        """
        Return the tmpdir that's accessible by VMs on system libvirt URIs
        """
        if guest.conn.is_xen():
            return "/var/lib/xen"
        return "/var/lib/libvirt/boot"

    @staticmethod
    def make_scratchdir(guest):
        """
        Determine the scratchdir for this URI, create it if necessary.
        scratchdir is the directory that's accessible by VMs
        """
        user_scratchdir = os.path.join(guest.conn.get_app_cache_dir(), "boot")
        system_scratchdir = InstallerTreeMedia.get_system_scratchdir(guest)

        # If we are a session URI, or we don't have access to the system
        # scratchdir, make sure the session scratchdir exists and use that.
        if (
            guest.conn.is_unprivileged()
            or not os.path.exists(system_scratchdir)
            or not os.access(system_scratchdir, os.W_OK)
        ):
            os.makedirs(user_scratchdir, 0o751, exist_ok=True)
            return user_scratchdir

        return system_scratchdir  # pragma: no cover

    def __init__(
        self,
        conn,
        location,
        location_kernel,
        location_initrd,
        install_kernel,
        install_initrd,
        install_kernel_args,
    ):
        self.conn = conn
        self.location = location
        self._location_kernel = location_kernel
        self._location_initrd = location_initrd
        self._install_kernel = install_kernel
        self._install_initrd = install_initrd
        self._install_kernel_args = install_kernel_args
        self._initrd_injections = []
        self._extra_args = []

        if location_kernel or location_initrd:
            if not location:
                raise ValueError(
                    _("location kernel/initrd may only be specified with a location URL/path")
                )
            if not (location_kernel and location_initrd):
                raise ValueError(_("location kernel/initrd must be specified as a pair"))

        self._cached_fetcher = None
        self._cached_data = None

        self._tmpfiles = []

        if self._install_kernel or self._install_initrd:
            self._media_type = MEDIA_KERNEL
        elif (
            not self.conn.is_remote()
            and os.path.exists(self.location)
            and os.path.isdir(self.location)
        ):
            self.location = os.path.abspath(self.location)
            self._media_type = MEDIA_DIR
        elif _is_url(self.location):
            self._media_type = MEDIA_URL
        else:
            self._media_type = MEDIA_ISO

        if (
            self.conn.is_remote()
            and not self._media_type == MEDIA_URL
            and not self._media_type == MEDIA_KERNEL
        ):
            raise ValueError(
                _("Cannot access install tree on remote connection: %s") % self.location
            )

        if self._media_type == MEDIA_ISO:
            InstallerTreeMedia.validate_path(self.conn, self.location)

    ########################
    # Install preparations #
    ########################

    def _get_fetcher(self, guest, meter):
        meter = progress.ensure_meter(meter)

        if not self._cached_fetcher:
            scratchdir = InstallerTreeMedia.make_scratchdir(guest)

            if self._media_type == MEDIA_KERNEL:
                self._cached_fetcher = urlfetcher.DirectFetcher(None, scratchdir, meter)
            else:
                self._cached_fetcher = urlfetcher.fetcherForURI(self.location, scratchdir, meter)

        self._cached_fetcher.meter = meter
        return self._cached_fetcher

    def _get_cached_data(self, guest, fetcher):
        if self._cached_data:
            return self._cached_data

        store = None
        osinfo = None
        os_media = None
        os_tree = None
        kernel_paths = []
        has_location_kernel = bool(self._location_kernel and self._location_initrd)

        if self._media_type == MEDIA_KERNEL:
            kernel_paths = [(self._install_kernel, self._install_initrd)]
        else:
            store = urldetect.getDistroStore(guest, fetcher, skip_error=has_location_kernel)

        if store:
            kernel_paths = store.get_kernel_paths()
            osinfo = store.get_osdict_info()
            os_media = store.get_os_media()
            os_tree = store.get_os_tree()
        if has_location_kernel:
            kernel_paths = [(self._location_kernel, self._location_initrd)]

        self._cached_data = _LocationData(osinfo, kernel_paths, os_media, os_tree)
        return self._cached_data

    def _prepare_kernel_url(self, guest, cache, fetcher):
        ignore = guest

        def _check_kernel_pairs():
            for kpath, ipath in cache.kernel_pairs:
                if fetcher.hasFile(kpath) and fetcher.hasFile(ipath):
                    return kpath, ipath
            raise RuntimeError(_("Couldn't find kernel for install tree."))  # pragma: no cover

        kernelpath, initrdpath = _check_kernel_pairs()
        kernel = fetcher.acquireFile(kernelpath)
        self._tmpfiles.append(kernel)
        initrd = fetcher.acquireFile(initrdpath)
        self._tmpfiles.append(initrd)

        perform_initrd_injections(initrd, self._initrd_injections, fetcher.scratchdir)

        return kernel, initrd

    ##############
    # Public API #
    ##############

    def _prepare_unattended_data(self, scripts):
        if not scripts:
            return

        for script in scripts:
            expected_filename = script.get_expected_filename()
            scriptpath = script.write()
            self._tmpfiles.append(scriptpath)
            self._initrd_injections.append((scriptpath, expected_filename))

    def _prepare_kernel_url_arg(self, guest, cache):
        osinfo = cache.osinfo or guest.osinfo.name
        osobj = OSDB.lookup_os(osinfo)
        return osobj.get_kernel_url_arg()

    def _prepare_kernel_args(self, guest, cache, unattended_scripts):
        install_args = None
        if unattended_scripts:
            args = []
            for unattended_script in unattended_scripts:
                cmdline = unattended_script.generate_cmdline()
                if cmdline:
                    args.append(cmdline)
            install_args = (" ").join(args)
            log.debug("Generated unattended cmdline: %s", install_args)
        elif self.is_network_url():
            kernel_url_arg = self._prepare_kernel_url_arg(guest, cache)
            if kernel_url_arg:
                install_args = "%s=%s" % (kernel_url_arg, self.location)

        if install_args:
            self._extra_args.append(install_args)

        if self._install_kernel_args:
            ret = self._install_kernel_args
        else:
            ret = " ".join(self._extra_args)

        if self._media_type == MEDIA_DIR and not ret:
            log.warning(
                _(
                    "Directory tree installs typically do not work "
                    "unless extra kernel args are passed to point the "
                    "installer at a network accessible install tree."
                )
            )
        return ret

    def prepare(self, guest, meter, unattended_scripts):
        fetcher = self._get_fetcher(guest, meter)
        cache = self._get_cached_data(guest, fetcher)

        self._prepare_unattended_data(unattended_scripts)
        kernel_args = self._prepare_kernel_args(guest, cache, unattended_scripts)

        kernel, initrd = self._prepare_kernel_url(guest, cache, fetcher)
        return kernel, initrd, kernel_args

    def cleanup(self, guest):
        ignore = guest
        for f in self._tmpfiles:
            log.debug("Removing %s", str(f))
            os.unlink(f)

        self._tmpfiles = []

    def set_initrd_injections(self, initrd_injections):
        self._initrd_injections = initrd_injections

    def set_extra_args(self, extra_args):
        self._extra_args = extra_args

    def cdrom_path(self):
        if self._media_type in [MEDIA_ISO]:
            return self.location

    def is_network_url(self):
        if self._media_type in [MEDIA_URL]:
            return self.location

    def detect_distro(self, guest):
        fetcher = self._get_fetcher(guest, None)
        cache = self._get_cached_data(guest, fetcher)
        return cache.osinfo

    def get_os_media(self, guest, meter):
        fetcher = self._get_fetcher(guest, meter)
        cache = self._get_cached_data(guest, fetcher)
        return cache.os_media

    def get_os_tree(self, guest, meter):
        fetcher = self._get_fetcher(guest, meter)
        cache = self._get_cached_data(guest, fetcher)
        return cache.os_tree

    def requires_internet(self, guest, meter):
        if self._media_type in [MEDIA_URL, MEDIA_DIR]:
            return True

        os_media = self.get_os_media(guest, meter)
        if os_media:
            return os_media.is_netinst()
        return False



# Copyright (C) 2018 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.


from .audio import DeviceAudio
from .char import DeviceChannel, DeviceConsole, DeviceParallel, DeviceSerial
from .controller import DeviceController
from .device import Device
from .disk import DeviceDisk
from .filesystem import DeviceFilesystem
from .graphics import DeviceGraphics
from .hostdev import DeviceHostdev
from .input import DeviceInput
from .interface import DeviceInterface
from .iommu import DeviceIommu
from .memballoon import DeviceMemballoon
from .memory import DeviceMemory
from .panic import DevicePanic
from .redirdev import DeviceRedirdev
from .rng import DeviceRng
from .shmem import DeviceShMem
from .smartcard import DeviceSmartcard
from .sound import DeviceSound
from .tpm import DeviceTpm
from .video import DeviceVideo
from .vsock import DeviceVsock
from .watchdog import DeviceWatchdog
from .pstore import DevicePstore


__all__ = [local for local in locals() if local.startswith("Device")]



#
# Support for parsing libvirt's domcapabilities XML
#
# Copyright 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import re
import xml.etree.ElementTree as ET

import libvirt

from .domain import DomainCpu
from .logger import log
from .xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty


########################################
# Genering <enum> and <value> handling #
########################################


class _Value(XMLBuilder):
    XML_NAME = "value"
    value = XMLProperty(".")


class _HasValues(XMLBuilder):
    values = XMLChildProperty(_Value)

    def get_values(self):
        return [v.value for v in self.values]

    def has_value(self, val):
        return val in self.get_values()


class _Enum(_HasValues):
    XML_NAME = "enum"
    name = XMLProperty("./@name")


class _CapsBlock(_HasValues):
    supported = XMLProperty("./@supported", is_yesno=True)
    _supported_present = XMLProperty("./@supported")
    enums = XMLChildProperty(_Enum)

    @property
    def present(self):
        return self._supported_present is not None

    def enum_names(self):
        return [e.name for e in self.enums]

    def has_enum(self, name):
        return name in self.enum_names()

    def get_enum(self, name):
        for enum in self.enums:
            if enum.name == name:
                return enum
        # Didn't find a match. Could be talking to older libvirt, or
        # driver with incomplete info. Return a stub enum
        return _Enum(self.conn)


def _make_capsblock(xml_root_name):
    """
    Build a class object representing a list of <enum> in the XML. For
    example, domcapabilities may have a block like:

    <graphics supported='yes'>
      <enum name='type'>
        <value>sdl</value>
        <value>vnc</value>
        <value>spice</value>
      </enum>
    </graphics>

    To build a class that tracks that whole <graphics> block, call this
    like _make_capsblock("graphics")
    """

    class TmpClass(_CapsBlock):
        pass

    setattr(TmpClass, "XML_NAME", xml_root_name)
    return TmpClass


################################
# SEV launch security handling #
################################


class _SEV(XMLBuilder):
    XML_NAME = "sev"
    supported = XMLProperty("./@supported", is_yesno=True)
    maxESGuests = XMLProperty("./maxESGuests")


#############################
# Misc toplevel XML classes #
#############################


class _OS(_CapsBlock):
    XML_NAME = "os"
    loader = XMLChildProperty(_make_capsblock("loader"), is_single=True)


class _Devices(_CapsBlock):
    XML_NAME = "devices"
    hostdev = XMLChildProperty(_make_capsblock("hostdev"), is_single=True)
    disk = XMLChildProperty(_make_capsblock("disk"), is_single=True)
    video = XMLChildProperty(_make_capsblock("video"), is_single=True)
    graphics = XMLChildProperty(_make_capsblock("graphics"), is_single=True)
    tpm = XMLChildProperty(_make_capsblock("tpm"), is_single=True)
    filesystem = XMLChildProperty(_make_capsblock("filesystem"), is_single=True)
    redirdev = XMLChildProperty(_make_capsblock("redirdev"), is_single=True)
    channel = XMLChildProperty(_make_capsblock("channel"), is_single=True)
    panic = XMLChildProperty(_make_capsblock("panic"), is_single=True)


class _Features(_CapsBlock):
    XML_NAME = "features"
    gic = XMLChildProperty(_make_capsblock("gic"), is_single=True)
    sev = XMLChildProperty(_SEV, is_single=True)
    hyperv = XMLChildProperty(_make_capsblock("hyperv"), is_single=True)


class _MemoryBacking(_CapsBlock):
    XML_NAME = "memoryBacking"


###############
# CPU classes #
###############


class _CPUModel(XMLBuilder):
    XML_NAME = "model"
    model = XMLProperty(".")
    usable = XMLProperty("./@usable")
    fallback = XMLProperty("./@fallback")


class _CPUMode(_CapsBlock):
    XML_NAME = "mode"
    name = XMLProperty("./@name")

    models = XMLChildProperty(_CPUModel)

    def get_model(self, name):
        for model in self.models:
            if model.model == name:
                return model


class _CPU(XMLBuilder):
    XML_NAME = "cpu"
    modes = XMLChildProperty(_CPUMode)

    def get_mode(self, name):
        for mode in self.modes:
            if mode.name == name:
                return mode


#############################
# CPU flags/baseline helpers#
#############################


def _convert_mode_to_cpu(xml, arch):
    root = ET.fromstring(xml)
    root.tag = "cpu"
    root.attrib = {}
    aelement = ET.SubElement(root, "arch")
    aelement.text = arch
    return ET.tostring(root, encoding="unicode")


def _get_expanded_cpu(domcaps, mode):
    cpuXML = _convert_mode_to_cpu(mode.get_xml(), domcaps.arch)
    log.debug("Generated CPU XML for security flag baseline:\n%s", cpuXML)

    try:
        expandedXML = domcaps.conn.baselineHypervisorCPU(
            domcaps.path,
            domcaps.arch,
            domcaps.machine,
            domcaps.domain,
            [cpuXML],
            libvirt.VIR_CONNECT_BASELINE_CPU_EXPAND_FEATURES,
        )
    except (libvirt.libvirtError, AttributeError):
        expandedXML = domcaps.conn.baselineCPU(
            [cpuXML], libvirt.VIR_CONNECT_BASELINE_CPU_EXPAND_FEATURES
        )

    return DomainCpu(domcaps.conn, expandedXML)


def _lookup_cpu_security_features(domcaps):
    ret = []
    sec_features = ["spec-ctrl", "ssbd", "ibpb", "virt-ssbd", "md-clear"]

    for m in domcaps.cpu.modes:
        if m.name != "host-model" or not m.supported:
            continue  # pragma: no cover

        try:
            cpu = _get_expanded_cpu(domcaps, m)
        except libvirt.libvirtError as e:  # pragma: no cover
            log.warning(_("Failed to get expanded CPU XML: %s"), e)
            break

        for feature in cpu.features:
            if feature.name in sec_features:
                ret.append(feature.name)

    log.debug("Found host-model security features: %s", ret)
    return ret


#################################
# DomainCapabilities main class #
#################################


class DomainCapabilities(XMLBuilder):
    XML_NAME = "domainCapabilities"
    os = XMLChildProperty(_OS, is_single=True)
    cpu = XMLChildProperty(_CPU, is_single=True)
    devices = XMLChildProperty(_Devices, is_single=True)
    features = XMLChildProperty(_Features, is_single=True)
    memorybacking = XMLChildProperty(_MemoryBacking, is_single=True)

    arch = XMLProperty("./arch")
    domain = XMLProperty("./domain")
    machine = XMLProperty("./machine")
    path = XMLProperty("./path")

    ################
    # Init helpers #
    ################

    @staticmethod
    def build_from_params(conn, emulator, arch, machine, hvtype):
        xml = None
        if conn.support.conn_domain_capabilities():
            try:
                xml = conn.getDomainCapabilities(emulator, arch, machine, hvtype)
                log.debug(
                    "Fetched domain capabilities for (%s,%s,%s,%s): %s",
                    emulator,
                    arch,
                    machine,
                    hvtype,
                    xml,
                )
            except Exception:  # pragma: no cover
                log.debug("Error fetching domcapabilities XML", exc_info=True)

        if not xml:
            # If not supported, just use a stub object
            return DomainCapabilities(conn)
        return DomainCapabilities(conn, parsexml=xml)

    @staticmethod
    def build_from_guest(guest):
        return DomainCapabilities.build_from_params(
            guest.conn, guest.emulator, guest.os.arch, guest.os.machine, guest.type
        )

    #########################
    # UEFI/firmware methods #
    #########################

    # Mapping of UEFI binary names to their associated architectures. We
    # only use this info to do things automagically for the user, it shouldn't
    # validate anything the user explicitly enters.
    _uefi_arch_patterns = {
        "i686": [
            r".*edk2-i386-.*\.fd",  # upstream qemu
            r".*ovmf-ia32.*",  # fedora, gerd's firmware repo
        ],
        "x86_64": [
            r".*edk2-x86_64-.*\.fd",  # upstream qemu
            r".*OVMF_CODE\.fd",  # RHEL
            r".*ovmf-x64/OVMF.*\.fd",  # gerd's firmware repo
            r".*ovmf-x86_64-.*",  # SUSE
            r".*ovmf.*",
            ".*OVMF.*",  # generic attempt at a catchall
        ],
        "aarch64": [
            r".*AAVMF_CODE\.fd",  # RHEL
            r".*aarch64/QEMU_EFI.*",  # fedora, gerd's firmware repo
            r".*aarch64.*",  # generic attempt at a catchall
            r".*edk2-aarch64-code\.fd",  # upstream qemu
        ],
        "armv7l": [
            r".*AAVMF32_CODE\.fd",  # Debian qemu-efi-arm package
            r".*arm/QEMU_EFI.*",  # fedora, gerd's firmware repo
            r".*edk2-arm-code\.fd",  # upstream qemu
        ],
        "riscv64": [
            r".*RISCV_VIRT_CODE\..*",  # Fedora, Debian
            r".*riscv64.*",  # generic attempt at a catchall
        ],
        "loongarch64": [
            r".*loongarch64/QEMU_CODE\..*",  # Fedora
            r".*loongarch64.*",  # generic attempt at a catchall
        ],
    }

    def find_uefi_path_for_arch(self):
        """
        Search the loader paths for one that matches the passed arch
        """
        if not self.arch_can_uefi():
            return  # pragma: no cover

        firmware_files = [f.value for f in self.os.loader.values]
        if self.conn.is_bhyve():
            for firmware_file in firmware_files:
                if "BHYVE_UEFI.fd" in firmware_file:
                    return firmware_file
            return firmware_files and firmware_files[0] or None  # pragma: no cover

        patterns = self._uefi_arch_patterns.get(self.arch)
        for pattern in patterns:
            for path in firmware_files:
                if re.match(pattern, path):
                    return path

    def label_for_firmware_path(self, path):
        """
        Return a pretty label for passed path, based on if we know
        about it or not
        """
        if not path:
            if self.arch in ["i686", "x86_64"]:
                return _("BIOS")
            return _("Default")

        for arch, patterns in self._uefi_arch_patterns.items():
            for pattern in patterns:
                if re.match(pattern, path):
                    return _("UEFI %(arch)s: %(path)s") % {"arch": arch, "path": path}

        return _("Custom: %(path)s") % {"path": path}

    def arch_can_uefi(self):
        """
        Return True if we know how to setup UEFI for the passed arch
        """
        return self.arch in self._uefi_arch_patterns

    def supports_uefi_loader(self):
        """
        Return True if libvirt advertises support for UEFI loader
        """
        return self.os.loader.get_enum("readonly").has_value("yes")

    def supports_firmware_efi(self):
        return self.os.get_enum("firmware").has_value("efi")

    #######################
    # CPU support methods #
    #######################

    def supports_safe_host_model(self):
        """
        Return True if domcaps reports support for cpu mode=host-model.
        host-model in fact predates this support, however it wasn't
        general purpose safe prior to domcaps advertisement.
        """
        m = self.cpu.get_mode("host-model")
        return m and m.supported and m.models[0].fallback == "forbid"

    def supports_safe_host_passthrough(self):
        """
        Return True if host-passthrough is safe enough to use by default.
        We limit this to domcaps new enough to report whether host-passthrough
        is migratable or not, which also means libvirt is about new enough
        to not taint the VM for using host-passthrough
        """
        m = self.cpu.get_mode("host-passthrough")
        return m and m.supported and "on" in m.get_enum("hostPassthroughMigratable").get_values()

    def supports_maximum_cpu_mode(self):
        m = self.cpu.get_mode("maximum")
        return m and m.supported

    def get_cpu_models(self):
        models = []

        for m in self.cpu.modes:
            if m.name == "custom" and m.supported:
                for model in m.models:
                    if model.usable != "no":
                        models.append(model.model)

        return models

    _features = None

    def get_cpu_security_features(self):
        if self._features is None:
            self._features = _lookup_cpu_security_features(self) or []
        return self._features

    ########################
    # Misc support methods #
    ########################

    def supports_sev_launch_security(self, check_es=False):
        """
        Returns False if either libvirt doesn't advertise support for SEV at
        all (< libvirt-4.5.0) or if it explicitly advertises it as unsupported
        on the platform
        """
        if check_es:
            return bool(self.features.sev.supported and self.features.sev.maxESGuests)
        return bool(self.features.sev.supported)

    def supports_video_bochs(self):
        """
        Returns False if either libvirt or qemu do not have support to bochs
        video type.
        """
        return self.devices.video.get_enum("modelType").has_value("bochs")

    def supports_video_qxl(self):
        if not self.devices.video.has_enum("modelType"):
            # qxl long predates modelType in domcaps, so if it is missing,
            # use spice support as a rough value
            return self.supports_graphics_spice()
        return self.devices.video.get_enum("modelType").has_value("qxl")

    def supports_video_virtio(self):
        return self.devices.video.get_enum("modelType").has_value("virtio")

    def supports_tpm_emulator(self):
        """
        Returns False if either libvirt or qemu do not have support for
        emulating a TPM.
        """
        models = self.devices.tpm.get_enum("model").get_values()
        backends = self.devices.tpm.get_enum("backendModel").get_values()

        if self.arch == "armv7l" and models == ["tpm-tis"]:
            # libvirt as of 8.4.0 can advertise armv7l tpm-tis support,
            # but then explicitly rejects that config. If we see it,
            # assume TPM is not supported
            # https://gitlab.com/libvirt/libvirt/-/issues/329
            return False

        return len(models) > 0 and bool("emulator" in backends)

    def supports_graphics_spice(self):
        if not self.devices.graphics.supported:
            # domcaps is too old, or the driver doesn't advertise graphics
            # support. Use our pre-existing logic
            if not self.conn.is_qemu() and not self.conn.is_test():
                return False
            return self.conn.caps.host.cpu.arch in ["i686", "x86_64"]

        return self.devices.graphics.get_enum("type").has_value("spice")

    def supports_channel_spicevmc(self):
        """
        Return False if libvirt explicitly advertises no support for
        spice channel
        """
        if self.devices.channel.supported is None:
            # Follow the original behavior in case of talking to older
            # libvirt.
            return True

        return self.devices.channel.get_enum("type").has_value("spicevmc")

    def supports_redirdev_usb(self):
        """
        Return False if libvirt explicitly advertises no support for
        USB redirect
        """
        if self.devices.redirdev.supported is None:
            # Follow the original behavior in case of talking to older
            # libvirt.
            return True

        return self.devices.redirdev.get_enum("bus").has_value("usb")

    def supports_filesystem_virtiofs(self):
        """
        Return True if libvirt advertises support for virtiofs
        """
        return self.devices.filesystem.get_enum("driverType").has_value("virtiofs")

    def supports_memorybacking_memfd(self):
        """
        Return True if libvirt advertises support for memfd memory backend
        """
        return self.memorybacking.get_enum("sourceType").has_value("memfd")

    def supported_hyperv_features(self):
        """
        Return list of supported Hyper-V features.
        """
        if not self.features.hyperv.supported:
            return []

        return self.features.hyperv.get_enum("features").get_values()

    def supported_panic_models(self):
        """
        Return list of supported panic device models.
        """
        return self.devices.panic.get_enum("model").get_values()



#
# Copyright 2014 Fujitsu Limited.
# Chen Hanxiao <chenhanxiao at cn.fujitsu.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..xmlbuilder import XMLBuilder, XMLProperty


class DomainMemtune(XMLBuilder):
    """
    Class for generating <memtune> XML
    """

    XML_NAME = "memtune"
    _XML_PROP_ORDER = ["hard_limit", "soft_limit", "swap_hard_limit", "min_guarantee"]

    hard_limit = XMLProperty("./hard_limit", is_int=True)
    soft_limit = XMLProperty("./soft_limit", is_int=True)
    swap_hard_limit = XMLProperty("./swap_hard_limit", is_int=True)
    min_guarantee = XMLProperty("./min_guarantee", is_int=True)



# Copyright 2013-2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import sys

import libvirt

from . import cli
from .cli import fail, fail_conflicting, print_stdout, print_stderr
from .guest import Guest
from .logger import log
from . import xmlutil


###################
# Utility helpers #
###################


def prompt_yes_or_no(msg):
    while 1:
        printmsg = msg + " (y/n): "
        sys.stdout.write(printmsg)
        sys.stdout.flush()
        inp = sys.stdin.readline().lower().strip()

        if inp in ["y", "yes"]:
            return True
        elif inp in ["n", "no"]:
            return False
        else:
            print_stdout(_("Please enter 'yes' or 'no'."))


def get_diff(origxml, newxml):
    diff = xmlutil.diff(origxml, newxml, "Original XML", "Altered XML")

    if diff:
        log.debug("XML diff:\n%s", diff)
    else:
        log.debug("No XML diff, didn't generate any change.")
    return diff


def set_osinfo(guest, osinfo):
    if osinfo is None:
        return

    osdata = cli.parse_osinfo(osinfo)
    if osdata.get_name():
        guest.set_os_name(osdata.get_name())


def defined_xml_is_unchanged(conn, domain, original_xml):
    rawxml = cli.get_xmldesc(domain, inactive=True)
    new_xml = Guest(conn, parsexml=rawxml).get_xml()
    return new_xml == original_xml


##################
# Action parsing #
##################


class Action:
    """
    Helper class tracking one pair of
        XML ACTION (ex. --edit) and
        XML OPTION (ex. --disk)
    """

    def __init__(self, action_name, selector, parserclass, parservalue):
        # one of ["edit", "add-device", "remove-device", "build-xml"]
        self.action_name = action_name
        # ex. for `--edit 1` this is selector="1"
        self.selector = selector
        # ParserDisk, etc
        self.parserclass = parserclass
        # ex for `--disk path=/foo` this is "path=/foo"
        self.parservalue = parservalue

    @property
    def is_edit(self):
        return self.action_name == "edit"

    @property
    def is_add_device(self):
        return self.action_name == "add-device"

    @property
    def is_remove_device(self):
        return self.action_name == "remove-device"

    @property
    def is_build_xml(self):
        return self.action_name == "build-xml"


def validate_action(action, conn, options):
    if options.osinfo is not None:
        if action.is_edit:
            fail(_("--osinfo/--os-variant is not supported with --edit"))
        if action.is_remove_device:
            fail(_("--osinfo/--os-variant is not supported with --remove-device"))
        if action.is_build_xml:
            fail(_("--osinfo/--os-variant is not supported with --build-xml"))

    if not action.parserclass.guest_propname and action.is_build_xml:
        fail(
            _("--build-xml not supported for {cli_flag}").format(
                cli_flag=action.parserclass.cli_flag_name()
            )
        )

    stub_guest = Guest(conn)
    if not action.parserclass.prop_is_list(stub_guest):
        if action.is_remove_device:
            fail(
                _("Cannot use --remove-device with {cli_flag}").format(
                    cli_flag=action.parserclass.cli_flag_name()
                )
            )
        if action.is_add_device:
            fail(
                _("Cannot use --add-device with {cli_flag}").format(
                    cli_flag=action.parserclass.cli_flag_name()
                )
            )

    if options.update and not action.parserclass.guest_propname:
        fail(
            _("Don't know how to --update for {cli_flag}").format(
                cli_flag=action.parserclass.cli_flag_name()
            )
        )


def check_action_collision(options):
    collisions = []
    actions = ["edit", "add-device", "remove-device", "build-xml"]
    for cliname in actions:
        optname = cliname.replace("-", "_")
        value = getattr(options, optname)
        if value not in [False, -1]:
            collisions.append((cliname, value))

    if len(collisions) == 0:
        fail(_("One of %s must be specified.") % ", ".join(["--" + c for c in actions]))
    if len(collisions) > 1:
        fail(_("Conflicting options %s") % ", ".join(["--" + c[0] for c in collisions]))

    action_name, selector = collisions[0]
    return action_name, selector


def check_xmlopt_collision(options):
    collisions = []
    for parserclass in cli.VIRT_PARSERS:
        value = getattr(options, parserclass.cli_arg_name)
        if value:
            collisions.append((parserclass, value))

    if len(collisions) == 0:
        fail(_("No change specified."))
    if len(collisions) != 1:
        fail(
            _("Only one change operation may be specified (conflicting options %s)")
            % [c[0].cli_flag_name() for c in collisions]
        )

    parserclass, parservalue = collisions[0]
    return parserclass, parservalue


def parse_action(conn, options):
    # Ensure there wasn't more than one device/xml config option
    # specified. So reject '--disk X --network X'
    parserclass, parservalue = check_xmlopt_collision(options)

    # Ensure only one of these actions was specified
    #   --edit
    #   --remove-device
    #   --add-device
    #   --build-xml
    action_name, selector = check_action_collision(options)

    action = Action(action_name, selector, parserclass, parservalue)
    validate_action(action, conn, options)
    return action


################
# Change logic #
################


def _find_objects_to_edit(guest, action_name, editval, parserclass):
    objlist = xmlutil.listify(parserclass.lookup_prop(guest))
    idx = None

    if editval is None:
        idx = 1
    elif editval.isdigit() or editval.startswith("-") and editval[1:].isdigit():
        idx = int(editval)

    if idx is not None:
        # Edit device by index
        if idx == 0:
            fail(_("Invalid --edit option '%s'") % editval)

        if not objlist:
            fail(
                _("No {cli_flag} objects found in the XML").format(
                    cli_flag=parserclass.cli_flag_name()
                )
            )
        if len(objlist) < abs(idx):
            fail(
                ngettext(
                    "'--edit {number}' requested but there's only "
                    "{maxnum} {cli_flag} object in the XML",
                    "'--edit {number}' requested but there are only "
                    "{maxnum} {cli_flag} objects in the XML",
                    len(objlist),
                ).format(number=idx, maxnum=len(objlist), cli_flag=parserclass.cli_flag_name())
            )

        if idx > 0:
            idx -= 1
        inst = objlist[idx]

    elif editval == "all":
        # Edit 'all' devices
        inst = objlist[:]

    else:
        # Lookup device by the passed prop string
        parserobj = parserclass(editval, guest=guest)
        inst = parserobj.lookup_child_from_option_string()
        if not inst:
            fail(_("No matching objects found for %s") % ("--%s %s" % (action_name, editval)))

    return inst


def action_edit(action, guest):
    parserclass = action.parserclass
    parservalue = action.parservalue
    selector = action.selector

    if parserclass.guest_propname:
        inst = _find_objects_to_edit(guest, "edit", selector, parserclass)
    else:
        inst = guest
        if selector and selector != "1" and selector != "all":
            fail(
                _(
                    "'--edit {option}' doesn't make sense with "
                    "{cli_flag}, just use empty '--edit'"
                ).format(option=selector, cli_flag=parserclass.cli_flag_name())
            )

    devs = []
    for editinst in xmlutil.listify(inst):
        devs += cli.run_parser(guest, parserclass, parservalue, editinst=editinst)
    return devs


def action_add_device(action, guest, osinfo, input_devs):
    parserclass = action.parserclass
    parservalue = action.parservalue

    set_osinfo(guest, osinfo)

    if input_devs:
        for dev in input_devs:
            guest.add_device(dev)
        devs = input_devs
    else:
        devs = cli.run_parser(guest, parserclass, parservalue)
        for dev in devs:
            dev.set_defaults(guest)

    return devs


def action_remove_device(action, guest):
    parserclass = action.parserclass
    parservalue = action.parservalue[-1]

    devs = _find_objects_to_edit(guest, "remove-device", parservalue, parserclass)
    devs = xmlutil.listify(devs)

    for dev in devs:
        guest.remove_device(dev)
    return devs


def action_build_xml(action, guest):
    parserclass = action.parserclass
    parservalue = action.parservalue

    devs = cli.run_parser(guest, parserclass, parservalue)
    for dev in devs:
        dev.set_defaults(guest)
    return devs


def perform_action(action, guest, options, input_devs):
    if action.is_add_device:
        return action_add_device(action, guest, options.osinfo, input_devs)
    if action.is_remove_device:
        return action_remove_device(action, guest)
    if action.is_edit:
        return action_edit(action, guest)
    raise xmlutil.DevError(
        "perform_action() incorrectly called with action_name=%s" % action.action_name
    )


def setup_device(dev):
    if getattr(dev, "DEVICE_TYPE", None) != "disk":
        return

    log.debug("Doing setup for disk=%s", dev)
    dev.build_storage(cli.get_meter())


def define_changes(conn, inactive_xmlobj, devs, action, confirm):
    if confirm:
        if not prompt_yes_or_no(_("Define '%s' with the changed XML?") % inactive_xmlobj.name):
            return False

    if action.is_add_device:
        for dev in devs:
            setup_device(dev)

    dom = conn.defineXML(inactive_xmlobj.get_xml())
    print_stdout(_("Domain '%s' defined successfully.") % inactive_xmlobj.name)
    return dom


def start_domain_transient(conn, xmlobj, devs, action, confirm):
    if confirm:
        if not prompt_yes_or_no(_("Start '%s' with the changed XML?") % xmlobj.name):
            return False

    if action.is_add_device:
        for dev in devs:
            setup_device(dev)

    try:
        dom = conn.createXML(xmlobj.get_xml())
    except libvirt.libvirtError as e:
        fail(
            _("Failed starting domain '%(domain)s': %(error)s")
            % {
                "domain": xmlobj.name,
                "error": e,
            }
        )
    else:
        print_stdout(_("Domain '%s' started successfully.") % xmlobj.name)
        return dom


def update_changes(domain, devs, action, confirm):
    if action.is_add_device:
        msg_confirm = _("%(xml)s\n\nHotplug this device to the guest '%(domain)s'?")
        msg_success = _("Device hotplug successful.")
        msg_fail = _("Error attempting device hotplug: %(error)s")
    elif action.is_remove_device:
        msg_confirm = _("%(xml)s\n\nHotunplug this device from the guest '%(domain)s'?")
        msg_success = _("Device hotunplug successful.")
        msg_fail = _("Error attempting device hotunplug: %(error)s")
    elif action.is_edit:
        msg_confirm = _("%(xml)s\n\nUpdate this device for the guest '%(domain)s'?")
        msg_success = _("Device update successful.")
        msg_fail = _("Error attempting device update: %(error)s")
    else:
        raise xmlutil.DevError(
            "update_changes() incorrectly called with action=%s" % action.action_name
        )

    for dev in devs:
        xml = dev.get_xml()

        if confirm:
            msg = msg_confirm % {
                "xml": xml,
                "domain": domain.name(),
            }
            if not prompt_yes_or_no(msg):
                continue

        if action.is_add_device:
            setup_device(dev)

        try:
            if action.is_add_device:
                domain.attachDeviceFlags(xml, libvirt.VIR_DOMAIN_AFFECT_LIVE)
            elif action.is_remove_device:
                domain.detachDeviceFlags(xml, libvirt.VIR_DOMAIN_AFFECT_LIVE)
            elif action.is_edit:
                domain.updateDeviceFlags(xml, libvirt.VIR_DOMAIN_AFFECT_LIVE)
        except libvirt.libvirtError as e:
            if "VIRTXML_TESTSUITE_UPDATE_IGNORE_FAIL" not in os.environ:
                fail(msg_fail % {"error": e})

        print_stdout(msg_success)
        if confirm:
            print_stdout("")


def prepare_changes(orig_xmlobj, options, action, input_devs=None):
    """
    Perform requested XML edits locally, but don't submit them to libvirt.
    Optionally perform any XML printing per user request

    :returns: (list of device objects, altered xmlobj)
    """
    origxml = orig_xmlobj.get_xml()
    xmlobj = orig_xmlobj.__class__(conn=orig_xmlobj.conn, parsexml=origxml)

    devs = perform_action(action, xmlobj, options, input_devs)
    newxml = xmlobj.get_xml()
    diff = get_diff(origxml, newxml)

    if not diff:
        log.warning(_("No XML diff was generated. The requested changes will have no effect."))

    if options.print_diff:
        if diff:
            print_stdout(diff)
    elif options.print_xml:
        print_stdout(newxml)

    return devs, xmlobj


#######################
# CLI option handling #
#######################


def parse_args():
    parser = cli.setupParser(
        "%(prog)s [options]",
        _("Edit libvirt XML using command line options."),
        introspection_epilog=True,
    )

    cli.add_connect_option(parser, "virt-xml")

    parser.add_argument("domain", nargs="?", help=_("Domain name, id, or uuid"))

    actg = parser.add_argument_group(_("XML actions"))
    actg.add_argument(
        "--edit",
        nargs="?",
        default=-1,
        help=_(
            "Edit VM XML. Examples:\n"
            "--edit --disk ...     (edit first disk device)\n"
            "--edit 2 --disk ...   (edit second disk device)\n"
            "--edit all --disk ... (edit all disk devices)\n"
            "--edit target=hda --disk ... (edit disk 'hda')\n"
        ),
    )
    actg.add_argument(
        "--remove-device",
        action="store_true",
        help=_(
            "Remove specified device. Examples:\n"
            "--remove-device --disk 1 (remove first disk)\n"
            "--remove-device --disk all (remove all disks)\n"
            "--remove-device --disk /some/path"
        ),
    )
    actg.add_argument(
        "--add-device",
        action="store_true",
        help=_("Add specified device. Example:\n--add-device --disk ..."),
    )
    actg.add_argument(
        "--build-xml",
        action="store_true",
        help=_(
            "Output built device XML. Domain is optional but "
            "recommended to ensure optimal defaults."
        ),
    )

    outg = parser.add_argument_group(_("Output options"))
    outg.add_argument(
        "--update",
        action="store_true",
        help=_(
            "Apply changes to the running VM.\n"
            "With --add-device, this is a hotplug operation.\n"
            "With --remove-device, this is a hotunplug operation.\n"
            "With --edit, this is an update device operation."
        ),
    )
    define_g = outg.add_mutually_exclusive_group()
    define_g.add_argument(
        "--define",
        action="store_true",
        help=_("Force defining the domain. Only required if a --print option was specified."),
    )
    define_g.add_argument(
        "--no-define", dest="define", action="store_false", help=_("Force not defining the domain.")
    )
    define_g.set_defaults(define=None)
    outg.add_argument("--start", action="store_true", help=_("Start the domain."))
    outg.add_argument(
        "--print-diff",
        action="store_true",
        help=_("Only print the requested change, in diff format"),
    )
    outg.add_argument(
        "--print-xml",
        action="store_true",
        help=_("Only print the requested change, in full XML format"),
    )
    outg.add_argument(
        "--confirm", action="store_true", help=_("Require confirmation before saving any results.")
    )

    cli.add_osinfo_option(parser, virtinstall=False)

    conv = parser.add_argument_group(_("Conversion options"))
    cli.ParserConvertToQ35.register()
    conv.add_argument(
        "--convert-to-q35",
        nargs="?",
        const=cli.VirtCLIParser.OPTSTR_EMPTY,
        help=_("Convert an existing VM from PC/i440FX to Q35."),
    )

    cli.ParserConvertToVNC.register()
    conv.add_argument(
        "--convert-to-vnc",
        nargs="?",
        const=cli.VirtCLIParser.OPTSTR_EMPTY,
        help=_(
            "Convert an existing VM to use VNC graphics. "
            "This removes any remnants of Spice graphics."
        ),
    )

    g = parser.add_argument_group(_("XML options"))
    cli.add_disk_option(g, editexample=True)
    cli.add_net_option(g)
    cli.add_gfx_option(g)
    cli.add_metadata_option(g)
    cli.add_memory_option(g)
    cli.vcpu_cli_options(g, editexample=True)
    cli.add_xml_option(g)
    cli.add_guest_xml_options(g)
    cli.add_boot_options(g)
    cli.add_device_options(g)

    misc = parser.add_argument_group(_("Miscellaneous Options"))
    cli.add_misc_options(misc, prompt=False, printxml=False, dryrun=False)

    cli.autocomplete(parser)

    return parser.parse_args()


###################
# main() handling #
###################


def main(conn=None):
    cli.earlyLogging()
    options = parse_args()

    if options.confirm or options.print_xml or options.print_diff or options.build_xml:
        options.quiet = False
    cli.setupLogging("virt-xml", options.debug, options.quiet)

    if cli.check_option_introspection(options):
        return 0
    if cli.check_osinfo_list(options):
        return 0

    options.stdinxml = None
    if not options.domain and not options.build_xml:
        if not sys.stdin.closed and not sys.stdin.isatty():
            if options.confirm:
                fail(_("Can't use --confirm with stdin input."))
            if options.update:
                fail(_("Can't use --update with stdin input."))
            options.stdinxml = sys.stdin.read()
        else:
            fail(_("A domain must be specified"))

    # Default to --define, unless:
    #  --no-define explicitly specified
    #  --print-* option is used
    #  XML input came from stdin
    if not options.print_xml and not options.print_diff:
        if options.stdinxml:
            if not options.define:
                options.print_xml = True
        else:
            if options.define is None:
                options.define = True
    if options.confirm and not options.print_xml:
        options.print_diff = True

    conn = cli.getConnection(options.connect, conn)
    action = parse_action(conn, options)

    domain = None
    active_xmlobj = None
    inactive_xmlobj = None
    if options.domain:
        domain, inactive_xmlobj, active_xmlobj = cli.get_domain_and_guest(conn, options.domain)
    else:
        inactive_xmlobj = Guest(conn, parsexml=options.stdinxml)
    vm_is_running = bool(active_xmlobj)

    if action.is_build_xml:
        built_devs = action_build_xml(action, inactive_xmlobj)
        for dev in built_devs:
            # pylint: disable=no-member
            print_stdout(xmlutil.unindent_device_xml(dev.get_xml()))
        return 0

    input_devs = None
    performed_update = False
    if options.update:
        if options.update and options.start:
            fail_conflicting("--update", "--start")
        if vm_is_running:
            input_devs, dummy = prepare_changes(active_xmlobj, options, action)
            update_changes(domain, input_devs, action, options.confirm)
            performed_update = True
        else:
            log.warning(_("The VM is not running, --update is inapplicable."))
        if not options.define:
            # --update and --no-define passed, so we are done
            return 0

    original_xml = inactive_xmlobj.get_xml()
    devs, xmlobj_to_define = prepare_changes(
        inactive_xmlobj, options, action, input_devs=input_devs
    )
    if not options.define:
        if options.start:
            start_domain_transient(conn, xmlobj_to_define, devs, action, options.confirm)
        return 0

    dom = define_changes(conn, xmlobj_to_define, devs, action, options.confirm)
    if not dom:
        # --confirm user said 'no'
        return 0

    if options.start:
        try:
            dom.create()
        except libvirt.libvirtError as e:  # pragma: no cover
            fail(
                _("Failed starting domain '%(domain)s': %(error)s")
                % {
                    "domain": inactive_xmlobj.name,
                    "error": e,
                }
            )
        print_stdout(_("Domain '%s' started successfully.") % inactive_xmlobj.name)

    elif vm_is_running and not performed_update:
        print_stdout(_("Changes will take effect after the domain is fully powered off."))
    elif defined_xml_is_unchanged(conn, dom, original_xml):
        log.warning(
            _(
                "XML did not change after domain define. You may "
                "have changed a value that libvirt is setting by default."
            )
        )

    return 0


def runcli():  # pragma: no cover
    try:
        sys.exit(main())
    except SystemExit as sys_e:
        sys.exit(sys_e.code)
    except KeyboardInterrupt:
        log.debug("", exc_info=True)
        print_stderr(_("Aborted at user request"))
    except Exception as main_e:
        fail(main_e)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.


from .blkiotune import DomainBlkiotune
from .clock import DomainClock
from .cpu import DomainCpu
from .cputune import DomainCputune
from .features import DomainFeatures
from .idmap import DomainIdmap
from .metadata import DomainMetadata
from .memorybacking import DomainMemoryBacking
from .memtune import DomainMemtune
from .numatune import DomainNumatune
from .os import DomainOs
from .pm import DomainPm
from .resource import DomainResource
from .seclabel import DomainSeclabel
from .keywrap import DomainKeyWrap
from .sysinfo import DomainSysinfo
from .vcpus import DomainVCPUs
from .xmlnsqemu import DomainXMLNSQemu
from .launch_security import DomainLaunchSecurity

__all__ = [local for local in locals() if local.startswith("Domain")]
